# Copyright 2024 David Trimm
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pandas as pd
import polars as pl
import numpy as np
import hashlib
import random
import urllib
import html
import os

from svglib.svglib import svg2rlg
from reportlab.graphics import renderPM
from PIL import Image
import io

from math import cos,sin,pi

from IPython.core import display as ipc_display

from IPython.display import Javascript, HTML, display

from rt_annotations_mixin       import RTAnnotationsMixin
from rt_art_mixin               import RTArtMixin
from rt_boxplot_mixin           import RTBoxplotMixin
from rt_calendarheatmap_mixin   import RTCalendarHeatmapMixin
from rt_chord_diagram_mixin     import RTChordDiagramMixin
from rt_choroplethmap_mixin     import RTChoroplethMapMixin
from rt_color_manager           import RTColorManager
from rt_datamanip_mixin         import RTDataManipMixin
from rt_geomaps_mixin           import RTGeoMapsMixin
from rt_geometry_mixin          import RTGeometryMixin
from rt_graph_layouts_mixin     import RTGraphLayoutsMixin
from rt_histogram_mixin         import RTHistogramMixin
from rt_json_mixin              import RTJSONMixin
from rt_layouts_mixin           import RTLayoutsMixin
from rt_link_mixin              import RTLinkMixin
from rt_linknode_mixin          import RTLinkNodeMixin
from rt_linknodeshortest_mixin  import RTLinkNodeShortestMixin
from rt_ontologies_mixin        import RTOntologiesMixin
from rt_panel_mixin             import RTPanelMixin
from rt_periodic_barchart_mixin import RTPeriodicBarChartMixin
from rt_piechart_mixin          import RTPieChartMixin
from rt_shapes_mixin            import RTShapesMixin
from rt_small_multiples_mixin   import RTSmallMultiplesMixin
from rt_temporal_barchart_mixin import RTTemporalBarChartMixin
from rt_text_mixin              import RTTextMixin
from rt_timeline_mixin          import RTTimelineMixin
from rt_wordcloud_mixin         import RTWordCloudMixin
from rt_xy_mixin                import RTXYMixin

__name__ = 'rtsvg'

class RACETrack(RTAnnotationsMixin,
                RTArtMixin,
                RTBoxplotMixin,
                RTCalendarHeatmapMixin,
                RTChordDiagramMixin,
                RTChoroplethMapMixin,
                RTDataManipMixin,
                RTGeoMapsMixin,
                RTGeometryMixin,
                RTGraphLayoutsMixin,
                RTHistogramMixin,
                RTJSONMixin,
                RTLayoutsMixin,   
                RTLinkMixin,             
                RTLinkNodeMixin,
                RTLinkNodeShortestMixin,
                RTOntologiesMixin,
                RTPanelMixin,
                RTPeriodicBarChartMixin,
                RTPieChartMixin,
                RTShapesMixin,
                RTSmallMultiplesMixin,
                RTTemporalBarChartMixin,
                RTTextMixin,
                RTTimelineMixin,
                RTWordCloudMixin,
                RTXYMixin):
    #
    # Constructor (or whatever this is called in Python)
    #
    def __init__(self):
        # Visualization globals
        self.co_mgr            = RTColorManager(self)
        self.default_font      = "Times"
        self.fformat           = '0.2f' # label formatting
        
        # Field transformations
                                  #
                                  # Time-based transformations
                                  #
        self.transforms        = ['day_of_week',          # day of the week
                                  'day_of_week_hour',     # day of the week plus the hour of the day
                                  'year',                 # year
                                  'quarter',              # quarter
                                  'year_quarter',         # year and quarter
                                  'month',                # month
                                  'year_month',           # year and month
                                  'year_month_day',       # year, month, and day
                                  'year_month_day_hour',  # year, month, day, and hour
                                  'day',                  # day (of the month)
                                  'day_of_year',          # day of the year
                                  'day_of_year_hour',     # day of the year w/ hour
                                  'hour',                 # hour (of the day)
                                  'minute',               # minute (of the hour)
                                  'second',               # second (of the minute)
                                  #
                                  # Numeric transformations
                                  #
                                  'log_bins',             # log-based binning
                                  #
                                  # IP-based transformations
                                  #
                                  'ipv4_cidr_24',         # ipv4_cidr_24
                                  'ipv4_cidr_16',         # ipv4_cidr_16
                                  'ipv4_cidr_08'          # ipv4_cidr_08
                                  ]

        # Used for reflections
        self.widgets           = ['boxplot',
                                  'calendarHeatmap',
                                  'chordDiagram',
                                  'choroplethMap',
                                  'histogram',
                                  'linkNode',
                                  'panelControl',       # Control Panel for Panel Impl
                                  'periodicBarChart',
                                  'pieChart',
                                  'temporalBarChart',
                                  'xy']
        
        # Cache for converting strings to integers
        RACETrack.hashcode_lu  = {}
        
        # Inits for mixins...  probably a better way to do this...
        self.__annotations_mixin_init__()
        self.__panel_mixin_init__()
        self.__periodic_barchart_mixin_init__()
        self.__temporal_barchart_mixin_init__()
        self.__text_mixin_init__()

    #
    # Render the SVG as HTML and display it within a notebook
    #
    def displaySVG(self,_svg):
        if type(_svg) != str:
            _svg = _svg._repr_svg_()
        return display(HTML(_svg))

    #
    # Render the SVG as an Image and display it within a notebook
    # - Uses an in memory image buffer
    # - Image form should save processing power for complicated SVGs
    #
    def displaySVGAsImage(self, _svg):
        if type(_svg) != str:
            _svg = _svg._repr_svg_()
        b = io.BytesIO()
        renderPM.drawToFile(svg2rlg(io.StringIO(_svg)), b, 'PNG')
        return ipc_display.Image(data=b.getvalue(),format='png',embed=True)

    #
    # isPandas() - is this a pandas dataframe?
    #
    def isPandas(self, df):
        return type(df) == pd.core.frame.DataFrame

    #
    # isPolars() - is this a polars dataframe?
    #
    def isPolars(self, df):
        return type(df) == pl.dataframe.frame.DataFrame

    #
    # copyDataFrame() - copy/clone a dataframe
    #
    def copyDataFrame(self, df):
        if   self.isPandas(df):
            return df.copy()
        elif self.isPolars(df):
            return df.clone()
        else:
            raise Exception('copyDataFrame() - accepts only pandas or polars dataframes')

    #
    # flattenTuple() - flatten a tuple
    #   flattenTuple(('fm','to'))                ==> ('fm', 'to')
    #   flattenTuple(('fm','to','other'))        ==> ('fm', 'to', 'other')
    #   flattenTuple(('a', ('b','c',('d','e')))) ==> ('a', 'b', 'c', 'd', 'e')
    #
    def flattenTuple(self, _tuple_):
        _ls_ = []
        for x in _tuple_:
            if type(x) == tuple:
                _ls_.extend(self.flattenTuple(x))
            else:
                _ls_.append(x)
        return tuple(_ls_)

    #
    # concatDataFrames() - concatenate dataframes
    #
    def concatDataFrames(self, dfs):
        if   self.isPandas(dfs[0]):
            return pd.concat(dfs)
        elif self.isPolars(dfs[0]):
            # Which columns are in common? // issue is that each component modifies the dataframe for the rendering...
            commons = set(dfs[0].columns)
            for _df_ in dfs:
                commons = commons & set(_df_.columns)
            # Make each dataframe conform to the common columns
            dfs_new = []
            for _df_ in dfs:
                dfs_new.append(_df_.drop(set(_df_.columns) - commons))
            # Perform the concatenate
            return pl.concat(dfs_new)
        else:
            raise Exception('concatDataFrames() - accepts only pandas or polars dataframes')

    #
    # createConcatColumn() - concatenate multiple columns together into a single column
    #
    def createConcatColumn(self, df, columns, new_column):
        def catFields(x, flds):
            s = str(x[flds[0]])
            for i in range(1,len(flds)):
                s += '|' + str(x[flds[i]])
            return s
        if self.isPandas(df):
            df[new_column] = df.apply(lambda x: catFields(x, columns), axis=1)
        elif self.isPolars(df):
            to_concat_new, str_casts = [], []
            for x in columns:
                if df[x].dtype != pl.String:
                    str_casts.append(pl.col(x).cast(str).alias('__' + x + '_as_str__'))
                    to_concat_new.append(pl.col('__' + x + '_as_str__'))
                else:
                    to_concat_new.append(pl.col(x))
            df = df.with_columns(*str_casts).with_columns(pl.concat_str(to_concat_new, separator='|').alias(new_column))
        else:
            raise Exception('createConcatColumn() - only pandas and polars supported')
        return df

    #
    # columnsAreTimestamps()
    # - Helper method to make a column into a suitable timestamp column
    # - ... because apparently the way to make a column into a type continues to evolve :(
    # - "format" parameter only applies to the polars typing method
    #
    def columnsAreTimestamps(self, df, columns, format=None):
        if type(columns) != list:
            columns = [columns]
        for _column_ in columns:
            if   self.isPandas(df):
                try:
                    df[_column_] = df[_column_].astype('datetime64[ms]')
                except:
                    df[_column_] = df[_column_].apply(lambda x: pd.to_datetime(x, utc=True).tz_convert(None))
            elif self.isPolars(df):
                try:
                    _format_ = format
                    if _format_ is None:
                        _format_ = self.guessTimestampFormat(str(df[0][_column_][0]))
                    df = df.with_columns(pl.col(_column_).str.strptime(pl.Datetime, format=_format_).cast(pl.Datetime))
                except:
                    df = df.with_columns(pl.col(_column_).map_elements(lambda x: pd.to_datetime(x, utc=True).tz_convert(None), return_dtype=pl.Datetime))
                    #as_series = df[_column_].map_elements(lambda x: pd.to_datetime(x, utc=True).tz_convert(None), return_dtype=pl.Datetime)
                    #_format_  = self.guessTimestampFormat(str(as_series[0]))
                    #df = df.drop(_column_)
                    #df = df.with_columns(as_series.str.strptime(pl.Datetime, format=_format_).cast(pl.Datetime))
            else:
                raise Exception('columnsAreTimestamps() - not a pandas or polars dataframe')
        return df

    #
    # guessTimestampFormat()
    # ... warning: bad code ahead :(
    # ... i hate timestamps
    #
    def guessTimestampFormat(self, sample):
        if (sample[-1] == 'z' or sample[-1] == 'Z') and ('t' in sample or 'T' in sample):
            if    len(sample) >= 22:
                return "%Y-%m-%dT%H:%M:%S.%fZ"
            elif  len(sample) == 20:
                return "%Y-%m-%dT%H:%M:%SZ"
            elif  len(sample) == 17: # Does this really occur?
                return "%Y-%m-%dT%H:%MZ"
            else:
                raise Exception(f'guessTimestampFormat() - no format specified for sample "{sample}" [1]')
        elif 't' in sample or 'T' in sample:
            if    len(sample) >  23:
                return "%Y-%m-%dT%H:%M:%S.%f"
            elif  len(sample) == 23:
                return "%Y-%m-%dT%H:%M:%S.%f"
            elif  len(sample) == 19:
                return "%Y-%m-%dT%H:%M:%S"
            elif  len(sample) == 16:
                return "%Y-%m-%dT%H:%M"
            elif  len(sample) == 13: # Doesn't seem to work in polars... wants both hour and minute
                return "%Y-%m-%dT%H"
            else:
                raise Exception(f'guessTimestampFormat() - no format specified for sample "{sample}" [2]')
        elif ' ' in sample:
            if    len(sample) >= 20:
                return "%Y-%m-%d %H:%M:%S.%f"
            elif  len(sample) == 19:
                return "%Y-%m-%d %H:%M:%S"        
            elif  len(sample) == 19:
                return "%Y-%m-%d %H:%M:%S"
            elif  len(sample) == 16:
                return "%Y-%m-%d %H:%M"
            elif  len(sample) == 13:  # Doesn't seem to work in polars... wants both hour and minute
                return "%Y-%m-%d %H"
            else:
                raise Exception(f'guessTimestampFormat() - no format specified for sample "{sample}" [3]')
        else:
            if    len(sample) == 10:
                return "%Y-%m-%d"
            elif  len(sample) == 7:
                return "%Y-%m"
            elif  len(sample) == 4:
                return "%Y"
            else:
                raise Exception(f'guessTimestampFormat() - no format specified for sample "{sample}" [4]')


    #
    # guessTimestampField() - guess the timestamp field
    #
    def guessTimestampField(self, df):
        if   self.isPandas(df):
            return self.__guessTimestampField_pandas__(df)
        elif self.isPolars(df):
            return self.__guessTimestampField_polars__(df)
        else:
            raise Exception(f'guessTimestampField() - only handles pandas and polars')
    def __guessTimestampField_pandas__(self, df):
        choices = df.select_dtypes(np.datetime64).columns
        if len(choices) == 1:
            return choices[0]
        elif len(choices) > 1:
            print('multiple timestamp fields... choosing the first (__guessTimestampField_pandas__)')
            return choices[0]
        else:
            raise Exception('no timestamp field supplied, cannot automatically determine field (__guessTimestampField_pandas__)')
    def __guessTimestampField_polars__(self, df):
        just_dt_columns = df.select(pl.col(pl.Datetime('us')))
        if len(just_dt_columns) == 0:
            just_dt_columns = df.select(pl.col(pl.Datetime('ns')))
        if len(just_dt_columns) == 0:
            just_dt_columns = df.select(pl.col(pl.Datetime('ms')))
        if   len(just_dt_columns.columns) > 1:
            print('multiple timestamp fields... choosing the first (__guessTimestampField_polars__)')
            return just_dt_columns.columns[0]
        elif len(just_dt_columns.columns) == 1:
            return just_dt_columns.columns[0]
        else:
            raise Exception('no timestamp field supplied, cannot automatically determine field (__guessTimestampField_polars__)')

    #
    # Return a consistent hashcode for a string
    #
    def hashcode(self,s):
        if type(s) != str: # Force non-strings to be strings
            s = str(s)
        if s not in RACETrack.hashcode_lu.keys(): # Cache the results so that we don't have to redo the calculation
            my_bytes = hashlib.sha256(s.encode('utf-8')).digest()
            value = ((my_bytes[0]<<24)&0x00ff000000) | ((my_bytes[1]<<16)&0x0000ff0000) | \
                    ((my_bytes[2]<< 8)&0x000000ff00) | ((my_bytes[3]<< 0)&0x00000000ff)
            RACETrack.hashcode_lu[s] = value
        return RACETrack.hashcode_lu[s]

    #
    # Encode a string into something safe for racetrack
    # ... in general, this code base uses pipes to separate strings... so it needs to be safe for that at least...
    #
    def stringEncode(self,s):
        return urllib.parse.quote_plus(s)
    
    #
    # Decode a string that was encoded with stringEncode()
    #
    def stringDecode(self,s):
        return urllib.parse.unquote_plus(s)

    #
    # Encode a string to make a valid SVG ID.
    # ... uses a colon escape sequence to encode any non [a-zA-Z0-9 ]
    #
    # From:  "https://www.dofactory.com/html/svg/id":
    #  "A unique alphanumeric string. The id value must begin with a letter ([A-Za-z]) and may be followed by 
    #   any number of letters, digits ([0-9]), hyphens (-), underscores (_), colons (:), and periods (.)."
    #
    def encSVGID(self, s):
        _enc = 'encsvgid_'
        if   type(s) == int:
            _enc += f'i_{s}'
            
        elif type(s) == str:        
            _enc += 's_'
            for c in s:
                if (c >= 'a' and c <= 'z') or \
                   (c >= 'A' and c <= 'Z') or \
                   (c >= '0' and c <= '9'):
                   _enc += c
                elif c == ' ':
                    _enc += '_'
                else:
                    as_int = ord(c)
                    _enc += ':'+str(as_int)+':'
        else:
            raise Exception('rtsvg.encSVGID() -- only strings and ints supported')
        return _enc

    #
    # Decode a string that was created by the encSVGID() method.
    #
    def decSVGID(self, s):
        if   s.startswith('encsvgid_i_'):
            s_prime = s[len('encsvgid_s_'):]
            return int(s_prime)
        elif s.startswith('encsvgid_s_'):
            s_prime = s[len('encsvgid_s_'):]
            _dec    = ''
            i = 0
            while i < len(s_prime):
                c = s_prime[i]
                if (c >= 'a' and c <= 'z') or \
                   (c >= 'A' and c <= 'Z') or \
                   (c >= '0' and c <= '9'):
                    _dec += c
                    i += 1
                elif c == '_':
                    _dec += ' '
                    i += 1
                elif c == ':':
                    i += 1
                    int_str = ''
                    while i < len(s_prime) and s_prime[i] != ':':
                        int_str += s_prime[i]
                        i += 1
                    _dec += chr(int(int_str))
                    i += 1
                else:
                    raise Exception(f'decSVGID() - failed to decode "{s}"')
            return _dec
        else:
            raise Exception(f'decSVGID() - unknown encoding type ... should be "_i_" or "_s_" "{s}"')

    # ****************************************************************************************************************
    #
    # Transformation Section
    #
    # ****************************************************************************************************************

    #
    # Transform a list of fields
    # - only handles one level of nesting for lists
    #
    def transformFieldListAndDataFrame(self, df, field_list):
        # Perform the transforms
        new_field_list = []
        for x in field_list:
            if type(x) == list:
                new_list = []
                for y in x:
                    if self.isTField(y) and y not in df.columns:
                        df,new_y = self.applyTransform(df,y)
                        new_list.append(new_y)
                    else:
                        new_list.append(y)
                new_field_list.append(new_list)
            else:
                if self.isTField(x) and x not in df.columns:
                    df,new_x = self.applyTransform(df, x)
                    new_field_list.append(new_x)
                else:
                    new_field_list.append(x)
        return df, new_field_list

    #
    # Determine if a field is a tfield
    #
    def isTField(self,tfield):
        return tfield is not None and type(tfield) == str and tfield.startswith('|tr|')      
    
    #
    # Return the applicable field for this transformation field (tfiled)
    #
    def tFieldApplicableField(self,tfield):
        if self.isTField(tfield):
            return '|'.join(tfield.split('|')[3:])
        return None
        
    #
    # Apply a tranformation field (tfield) to a dataframe and return the new dataframe and the calculated new field
    # ... we'll want set-based counting -- so we'll make sure it's never just a number
    #
    def applyTransform(self, df, tfield):
        if tfield is not None and tfield.startswith('|tr|') and tfield not in df.columns:
            transform = tfield.split('|')[2]
            field     = '|'.join(tfield.split('|')[3:])

            ipv4CIDR08 = lambda ipv4: ipv4.split('.')[0]
            ipv4CIDR16 = lambda ipv4: ipv4.split('.')[0] + '.' + ipv4.split('.')[1]
            ipv4CIDR24 = lambda ipv4: ipv4.split('.')[0] + '.' + ipv4.split('.')[1] + '.' + ipv4.split('.')[2]

            #
            # PANDAS Version
            #
            if self.isPandas(df):            
                if   transform == 'day_of_week':
                    df[tfield] = df[field].apply(lambda x: str(x.day_name()[:3]))
                elif transform == 'day_of_week_hour':
                    df[tfield] = df[field].apply(lambda x: f'{x.day_name()[:3]}-{x.hour:02}')
                elif transform == 'year':
                    df[tfield] = df[field].apply(lambda x: str(x.year))
                elif transform == 'year_quarter':
                    df[tfield] = df[field].apply(lambda x: f'{x.year}Q{x.quarter}')
                elif transform == 'quarter':
                    df[tfield] = df[field].apply(lambda x: f'Q{x.quarter}')
                elif transform == 'month':
                    df[tfield] = df[field].apply(lambda x: x.month_name()[:3])
                elif transform == 'year_month':
                    df[tfield] = df[field].apply(lambda x: f'{x.year}-{x.month:02}')
                elif transform == 'year_month_day':
                    df[tfield] = df[field].apply(lambda x: f'{x.year}-{x.month:02}-{x.day:02}')
                elif transform == 'year_month_day_hour':
                    df[tfield] = df[field].apply(lambda x: f'{x.year}-{x.month:02}-{x.day:02} {x.hour:02}')
                elif transform == 'day':
                    df[tfield] = df[field].apply(lambda x: f'{x.day:02}')
                elif transform == 'day_of_year':
                    df[tfield] = df[field].apply(lambda x: f'{x.day_of_year:03}')
                elif transform == 'day_of_year_hour':
                    df[tfield] = df[field].apply(lambda x: f'{x.day_of_year:03}_{x.hour:02}')
                elif transform == 'hour':
                    df[tfield] = df[field].apply(lambda x: f'{x.hour:02}')
                elif transform == 'minute':
                    df[tfield] = df[field].apply(lambda x: f'{x.minute:02}')
                elif transform == 'second':
                    df[tfield] = df[field].apply(lambda x: f'{x.second:02}')
                elif transform == 'log_bins':
                    df[tfield] = df[field].apply(lambda x: self.transformLogBins(x))
                elif transform == 'ipv4_cidr_24': # ipv4_cidr_24
                    df[tfield] = df[field].apply(lambda x: ipv4CIDR24(x))
                elif transform == 'ipv4_cidr_16': # ipv4_cidr_16
                    df[tfield] = df[field].apply(lambda x: ipv4CIDR16(x))
                elif transform == 'ipv4_cidr_08': # ipv4_cidr_08
                    df[tfield] = df[field].apply(lambda x: ipv4CIDR08(x))

            #
            # POLARS Version
            #            
            elif self.isPolars(df):
                if   transform == 'day_of_week':
                    df = df.with_columns(pl.col(field).dt.strftime('%A').str.slice(0,3).alias(tfield))
                elif transform == 'day_of_week_hour':
                    _intermediate_dow_  = '__day_of_week_hour_dow__'
                    df = df.with_columns(pl.col(field).dt.strftime('%A').str.slice(0,3).alias(_intermediate_dow_))
                    _intermediate_hour_ = '__day_of_week_hour_hour__'
                    df = df.with_columns(pl.col(field).dt.strftime('-%H').alias(_intermediate_hour_))
                    df = df.with_columns(pl.concat_str(_intermediate_dow_, _intermediate_hour_).alias(tfield))
                elif transform == 'year':
                    # df = df.with_columns(pl.col(field).dt.strftime('%Y').cast(pl.Int64).alias(tfield))
                    df = df.with_columns(pl.col(field).dt.strftime('%Y').cast(pl.Int64).alias(tfield))
                elif transform == 'year_quarter':
                    _intermediate_quarter_ = '__quarter__'                    
                    df = df.with_columns(pl.col(field).dt.quarter().cast(str).alias(_intermediate_quarter_))
                    _intermediate_year_    = '__year__'
                    df = df.with_columns(pl.col(field).dt.strftime('%YQ').alias(_intermediate_year_))
                    df = df.with_columns(pl.concat_str(_intermediate_year_, _intermediate_quarter_).alias(tfield))
                elif transform == 'quarter':
                    _intermediate_quarter_ = '__quarter__'
                    df = df.with_columns(pl.col(field).dt.quarter().cast(str).alias(_intermediate_quarter_))
                    df = df.with_columns(pl.format('Q{}', _intermediate_quarter_).alias(tfield))
                elif transform == 'month':
                    df = df.with_columns(pl.col(field).dt.strftime('%h').alias(tfield))
                elif transform == 'year_month':
                    df = df.with_columns(pl.col(field).dt.strftime('%Y-%m').alias(tfield))
                elif transform == 'year_month_day':
                    df = df.with_columns(pl.col(field).dt.strftime('%Y-%m-%d').alias(tfield))
                elif transform == 'year_month_day_hour':
                    df = df.with_columns(pl.col(field).dt.strftime('%Y-%m-%d %H').alias(tfield))
                elif transform == 'day':
                    # df = df.with_columns(pl.col(field).dt.strftime('%d').cast(pl.Int64).alias(tfield))
                    df = df.with_columns(pl.col(field).dt.strftime('%d').alias(tfield))
                elif transform == 'day_of_year':
                    # df = df.with_columns(pl.col(field).dt.strftime('%j').cast(pl.Int64).alias(tfield))
                    df = df.with_columns(pl.col(field).dt.strftime('%j').alias(tfield))
                elif transform == 'day_of_year_hour':
                    df = df.with_columns(pl.col(field).dt.strftime('%j_%H').alias(tfield))
                elif transform == 'hour':
                    # df = df.with_columns(pl.col(field).dt.strftime('%H').cast(pl.Int64).alias(tfield))
                    df = df.with_columns(pl.col(field).dt.strftime('%H').alias(tfield))
                elif transform == 'minute':
                    # df = df.with_columns(pl.col(field).dt.strftime('%M').cast(pl.Int64).alias(tfield))
                    df = df.with_columns(pl.col(field).dt.strftime('%M').alias(tfield))
                elif transform == 'second':
                    # df = df.with_columns(pl.col(field).dt.strftime('%S').cast(pl.Int64).alias(tfield))
                    df = df.with_columns(pl.col(field).dt.strftime('%S').alias(tfield))
                elif transform == 'log_bins':
                    df = df.with_columns(pl.col(field).apply(lambda x: self.transformLogBins(x)).alias(tfield))
                elif transform == 'ipv4_cidr_24': # ipv4_cidr_24
                    df = df.with_columns(pl.col(field).map_elements(lambda x: ipv4CIDR24(x)).alias(tfield))
                elif transform == 'ipv4_cidr_16': # ipv4_cidr_16
                    df = df.with_columns(pl.col(field).map_elements(lambda x: ipv4CIDR16(x)).alias(tfield))
                elif transform == 'ipv4_cidr_08': # ipv4_cidr_08
                    df = df.with_columns(pl.col(field).map_elements(lambda x: ipv4CIDR08(x)).alias(tfield))
            else:
                raise Exception('applyTransform() - df is neither a pandas nor a polars dataframe')

        return df,tfield

    #
    # Define the natural order for the elements returned by a tfield
    # - a few degenerate cases exist -- for example, year_month_day if give many centuries...
    #
    def transformNaturalOrder(self, df, tfield):
        _order = []
        _order_dow     = ['Mon','Tue','Wed','Thu','Fri','Sat','Sun']
        _order_quarter = ['Q1','Q2','Q3','Q4']
        _order_month   = ['Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec']
        if tfield is not None and tfield.startswith('|tr|'):
            transform = tfield.split('|')[2]
            field     = '|'.join(tfield.split('|')[3:])
            if   transform == 'day_of_week':
                _order = _order_dow
            elif transform == 'day_of_week_hour':
                for _dow in _order_dow:
                    for _hour in range(0,24):
                        _order.append(f'{_dow}-{_hour:02}')
            elif transform == 'year':
                for _year in range(df[field].min().year, df[field].max().year+1):
                    _order.append(f'{_year}')
            elif transform == 'year_quarter':
                for _year in range(df[field].min().year, df[field].max().year+1):
                    for _quarter in range(1,5):
                        _order.append(f'{_year}Q{_quarter}')
            elif transform == 'quarter':
                _order = _order_quarter
            elif transform == 'month':
                _order = _order_month
            elif transform == 'year_month':
                for _date in pd.date_range(start=df[field].min(), end=df[field].max(), freq='M'):
                    _order.append(f'{_date.year}-{_date.month:02}')
            elif transform == 'year_month_day':
                for _date in pd.date_range(start=df[field].min(), end=df[field].max(), freq='D'):
                    _order.append(f'{_date.year}-{_date.month:02}-{_date.day:02}')
            elif transform == 'year_month_day_hour':
                for _date in pd.date_range(start=df[field].min(), end=df[field].max(), freq='H'):
                    _order.append(f'{_date.year}-{_date.month:02}-{_date.day:02} {_date.hour:02}')
            elif transform == 'day':
                for _day in range(1,32):
                    _order.append(f'{_day:02}')
            elif transform == 'day_of_year':
                if df[field].min().year == df[field].max().year:
                    _day_min, _day_max = pd.to_datetime(df[field].min()).day_of_year, pd.to_datetime(df[field].max()).day_of_year
                    for _day in range(_day_min,_day_max+1):
                        _order.append(f'{_day:03}')
                else:
                    for _day in range(1,366):
                        _order.append(f'{_day:03}')
            elif transform == 'day_of_year_hour':
                if df[field].min().year == df[field].max().year:
                    _day_min, _day_max = pd.to_datetime(df[field].min()).day_of_year, pd.to_datetime(df[field].max()).day_of_year                    
                    for _day in range(_day_min,_day_max+1):
                        for _hour in range(0,24):
                            _order.append(f'{_day:03}_{_hour:02}')
                else:
                    for _day in range(1,366):
                        for _hour in range(0,24):
                            _order.append(f'{_day:03}_{_hour:02}')
            elif transform == 'hour':
                for _hour in range(0,24):
                    _order.append(f'{_hour:02}')
            elif transform == 'minute' or transform == 'second':
                for _minute in range(0,60):
                    _order.append(f'{_minute:02}')
            elif transform == 'log_bins':
                _order = ['< 0', '= 0', '<= 1', '<= 10','<= 100', '<= 1K', '<= 100K', '<= 1M', '> 1M']
            elif transform == 'ipv4_cidr_24':
                _pad_ = lambda x: f'{int(x.split(".")[0]):03}.{int(x.split(".")[1]):03}.{int(x.split(".")[2]):03}'
                _order = list(set(df[tfield]))
                _order.sort(key=_pad_)
            elif transform == 'ipv4_cidr_16':
                _pad_ = lambda x: f'{int(x.split(".")[0]):03}.{int(x.split(".")[1]):03}'
                _order = list(set(df[tfield]))
                _order.sort(key=_pad_)
            elif transform == 'ipv4_cidr_08':
                _pad_ = lambda x: f'{int(x.split(".")[0]):03}'
                _order = list(set(df[tfield]))
                _order.sort(key=_pad_)
        return _order

    #
    # Create a tranformation field (tfield)
    #
    def createTField(self,field,trans):
        if trans in self.transforms:
            tfield = '|tr|'+trans+'|'+field
        else:
            raise Exception(f'Transform "{trans}" is not defined')
        return tfield

    #
    # Make simple log-based bins
    # - strings are equivalent to a color scheme in RTColorManager class.
    #
    def transformLogBins(self, x):
        x = float(x)
        if   x < 0:
            return '< 0'
        elif x == 0:
            return '= 0'
        elif x <= 1:
            return '<= 1'
        elif x <= 10:
            return '<= 10'
        elif x <= 100:
            return '<= 100'
        elif x <= 1000:
            return '<= 1K'
        elif x <= 100000:
            return '<= 100K'
        elif x <= 1000000:
            return '<= 1M'
        else:
            return '> 1M'

    #
    # Identify columns needed from widget parameters
    #
    def identifyColumnsFromParameters(self, param_name, param_lu, columns_set):
        # print(f'identifyColumnsFromParameters(,"{param_name}","{param_lu}","{columns_set}")') # DEBUG
        if param_name in param_lu.keys() and param_lu[param_name] is not None:
            v = param_lu[param_name]
            self.__recursiveDecompose__(v, columns_set)

    def __recursiveDecompose__(self, something, columns_set):
        if   type(something) == str:
            columns_set.add(something)
        elif type(something) == bool: # unclear why None may be converted to False // is that what's happening?
            pass # do nothing
        elif type(something) == list or type(something) == tuple:
            for x in something:
                self.__recursiveDecompose__(x, columns_set)
        elif type(something) == dict:
            pass # do nothing
        else:
            raise Exception(f'Unknown type ("{type(something)}") for ("{something}") encountered in identifyColumnsFromParameters()')

    #
    # polarsFilterColumnsWithNaNs() -- filter specified columns with NaN values
    #
    def polarsFilterColumnsWithNaNs(self, df, cols):
        _eval_ = []
        for col in cols:
            _eval_.append(f'(pl.col("{col}").is_not_null())')
        df = df.filter(eval('&'.join(_eval_)))
        return df

    #
    # polarsCounter() -- return a dataframe with fields and an __count__ column.
    #
    def polarsCounter(self, df, fields, count_by=None, count_by_set=False):
        fields = [fields] if type(fields) != list else fields
        if count_by is not None and count_by_set == False:
            if self.fieldIsArithmetic(df, count_by) == False:
                count_by_set = True
        if count_by is None:
            return df.group_by(fields).agg(pl.count().alias('__count__'))
        elif count_by_set and count_by in fields:
            df_min = df.drop(set(df.columns) - set(fields) - set([count_by]))
            df_dupe = df_min.with_columns(pl.col(count_by).alias('__count__'))
            return df_dupe.group_by(fields).n_unique()
        elif count_by_set:
            df_min = df.drop(set(df.columns) - set(fields) - set([count_by])).rename({count_by:'__count__'})
            return df_min.group_by(fields).n_unique()
        elif count_by in fields:
            df_min = df.drop(set(df.columns) - set(fields) - set([count_by]))
            df_dupe = df_min.with_columns(pl.col(count_by).alias('__count__'))
            return df_dupe.group_by(fields).sum()
        else:
            df_min = df.drop(set(df.columns) - set(fields) - set([count_by])).rename({count_by:'__count__'})
            return df_min.group_by(fields).sum()

    #
    # Determine If A Column Has To Be Counted By Set Operation
    #
    def countBySet(self, 
                   df,         # dataframe
                   count_by):  # field to check
        if count_by is None:
            return False
        if type(df) == list:
            for _df in df:
                if self.isPandas(_df):
                    if count_by in _df.columns:
                        if _df[count_by].dtypes != np.int64   and \
                        _df[count_by].dtypes != np.int32   and \
                        _df[count_by].dtypes != np.float64 and \
                        _df[count_by].dtypes != np.float32:
                            return True
                elif self.isPolars(_df):
                    if _df[count_by].is_float() == False and _df[count_by].is_integer() == False:
                        return True
                else:
                    raise Exception('countBySet() - not a pandas or polars dataframe')
            return False
        else:
            if self.isPandas(df):
                return df[count_by].dtypes != np.int64   and \
                       df[count_by].dtypes != np.int32   and \
                       df[count_by].dtypes != np.float64 and \
                       df[count_by].dtypes != np.float32
            elif self.isPolars(df):
                return df[count_by].dtype.is_float() == False and df[count_by].dtype.is_integer() == False
            else:
                raise Exception('countBySet() - not a pandas or polars dataframe')
    
    #
    # fieldIsArithmetic()
    # ... determine if a field can be operated on by arithmetic
    # ... maybe the oposite of the above?
    #
    def fieldIsArithmetic(self, df, field):
        if self.isPandas(df):
            return df[field].dtypes == np.int64   or \
                   df[field].dtypes == np.int32   or \
                   df[field].dtypes == np.float64 or \
                   df[field].dtypes == np.float32
        elif self.isPolars(df):
            return df[field].dtype.is_float() or df[field].dtype.is_integer()
        else:
            raise Exception('fieldIsArithmetic() - not a pandas or polars dataframe')

    #
    # Determine color ordering based on quantity
    #
    def colorRenderOrder(self, 
                         df,                   # dataframe
                         color_by,             # color_by field
                         count_by,             # count_by field
                         count_by_set=False):  # for the field set, count by set operation
        # If no color, then return none...
        if color_by is None:
            return None
        
        # Make sure we can count by numeric summation
        if count_by_set == False:
            count_by_set = self.countBySet(df, count_by)
        
        if   self.isPandas(df):      
            return self.__colorQuantities_pandas__(df, color_by, count_by, count_by_set)
        elif self.isPolars(df):
            return self.__colorQuantities_polars__(df, color_by, count_by, count_by_set)
        else:
            raise Exception('colorRenderOrder() - not a pandas or polars dataframe')

    #
    # Determine color quantities
    #
    def __colorQuantities_pandas__(self, 
                                   df,            # dataframe 
                                   color_by,      # color_by field
                                   count_by,      # count_by field
                                   count_by_set): # for the field set, count by set operation
        # For count by set... when count_by == color by... then we'll count by rows
        if count_by is not None and count_by_set and count_by == color_by:
            count_by = None

        if count_by is None:
            return df.groupby(color_by).size().sort_values(ascending=False)
        elif count_by_set:
            _df = pd.DataFrame(df.groupby([color_by,count_by]).size()).reset_index()
            return _df.groupby(color_by).size().sort_values(ascending=False)
        elif count_by == color_by:
            _df = df.groupby(color_by).size().reset_index()
            _df['__mult__'] = _df.apply(lambda x: x[color_by]*x[0],axis=1)
            return _df.groupby(color_by)['__mult__'].sum().sort_values(ascending=False)
        else:
            return df.groupby(color_by)[count_by].sum().sort_values(ascending=False)

    #
    # Determine color quantities
    # ... polars has no concept of index... so this isn't going to be a 1-for-1 method...
    # ... i hope there isn't a columns named 'count_by' or 'color_by'
    #
    def __colorQuantities_polars__(self, 
                                   df,            # dataframe 
                                   color_by,      # color_by field
                                   count_by,      # count_by field
                                   count_by_set): # for the field set, count by set operation
        # For count by set... when count_by == color by... then we'll count by rows
        if count_by is not None and count_by_set and count_by == color_by:
            count_by = None

        # Remove all the columns that aren't used for the operation...
        # ... and rename them to 'count_by' and 'color_by'
        if count_by is not None:
            _to_drop_ = list(df.columns)
            _to_drop_.remove(color_by)
            if count_by in _to_drop_:
                _to_drop_.remove(count_by)
            df = df.drop(_to_drop_)
            if color_by != count_by:
                df = df.rename({color_by:'color_by', count_by:'count_by'})
                color_by = 'color_by'
                count_by = 'count_by'
            else:
                df = df.rename({color_by:'color_by'})
                color_by = count_by = 'color_by'
        else:
            _to_drop_ = list(df.columns)
            _to_drop_.remove(color_by)
            df = df.drop(_to_drop_)
            df = df.rename({color_by:'color_by'})
            color_by = 'color_by'

        if count_by is None:            
            _df_ = df.group_by(color_by).agg(pl.count()) # .sort('count', descending=True)
        elif count_by_set:
            _df_ = df.group_by(color_by).n_unique() # .sort(count_by,descending=True)
        elif count_by == color_by:
            _df_ = df.group_by(color_by).agg(pl.count())
            _df_ = _df_.with_columns(pl.col(color_by).mul(pl.col('count')).alias('__mult__'))
            _df_ = _df_.sort('__mult__',descending=True)
            _df_ = _df_.drop('count')
            _df_ = _df_.rename({'__mult__':'count'})
        else:
            _df_ = df.group_by(color_by).agg(pl.sum(count_by)) # .sort(count_by, descending=True)

        if 'count_by' in set(_df_.columns):
            _df_ = _df_.rename({'count_by':'count'})
        _df_ = _df_.rename({'color_by':'index'})

        _df_ = _df_.sort('index').sort('count', descending=True)
        
        return _df_
    
    #
    # Colorize Bar
    #
    def colorizeBar(self,
                    df,                  # dataframe
                    global_color_order,  # global color ordering - returned from colorRenderOrder()
                    color_by,            # color_by field
                    count_by,            # count_by field
                    count_by_set,        # for the field set, count by set operation
                    x,                   # x coordinate of the bar base
                    y,                   # y coordinate of the bar base
                    bar_len,             # total bar length -- for vertical, this is the height
                    bar_sz,              # size of bar -- for vertical, this is the width
                    horz):               # true for horizontal bars (histogram), false for vertical bars
        if   self.isPandas(df):
            return self.__colorizeBar_pandas__(df, global_color_order, color_by, count_by, count_by_set, x, y, bar_len, bar_sz, horz)
        elif self.isPolars(df):
            return self.__colorizeBar_polars__(df, global_color_order, color_by, count_by, count_by_set, x, y, bar_len, bar_sz, horz)
        else:
            raise Exception('colorizeBar() - not a pandas or polars dataframe')

    def __colorizeBar_pandas__(self,
                               df,                  # dataframe
                               global_color_order,  # global color ordering - returned from colorRenderOrder()
                               color_by,            # color_by field
                               count_by,            # count_by field
                               count_by_set,        # for the field set, count by set operation
                               x,                   # x coordinate of the bar base
                               y,                   # y coordinate of the bar base
                               bar_len,             # total bar length -- for vertical, this is the height
                               bar_sz,              # size of bar -- for vertical, this is the width
                               horz):               # true for horizontal bars (histogram), false for vertical bars
        svg = []
        if bar_len > 0:
            _co = self.co_mgr.getTVColor('data','default')
            # Default bar w/out color
            if horz:
                svg.append(f'<rect x="{x}" y="{y}" width="{bar_len}" height="{bar_sz}" fill="{_co}" />')
            else:
                svg.append(f'<rect x="{x}" y="{y-bar_len}" width="{bar_sz}" height="{bar_len}" fill="{_co}" />')
            # Colorize it
            if color_by is not None:
                quantities   = self.colorRenderOrder(df, color_by, count_by, count_by_set)
                value        = quantities.sum()
                quantities   = quantities[quantities > value/bar_len]
                intersection = self.__myIntersection__(global_color_order.index, quantities.index)
                d = x if horz else y
                for cb_bin in intersection:
                    v = quantities[cb_bin]
                    l = bar_len * v / value
                    if l >= 1.0:
                        _co = self.co_mgr.getColor(cb_bin)
                        if horz:
                            svg.append(f'<rect x="{d}" y="{y}" width="{l}" height="{bar_sz}" fill="{_co}" />')
                            d += l
                        else:
                            svg.append(f'<rect x="{x}" y="{d-l}" width="{bar_sz}" height="{l}" fill="{_co}" />')
                            d -= l
        return ''.join(svg)

    def __colorizeBar_polars__(self,
                               df,                  # dataframe
                               global_color_order,  # global color ordering - returned from colorRenderOrder()
                               color_by,            # color_by field
                               count_by,            # count_by field
                               count_by_set,        # for the field set, count by set operation
                               x,                   # x coordinate of the bar base
                               y,                   # y coordinate of the bar base
                               bar_len,             # total bar length -- for vertical, this is the height
                               bar_sz,              # size of bar -- for vertical, this is the width
                               horz):               # true for horizontal bars (histogram), false for vertical bars
        svg = []
        if bar_len > 0:
            _co = self.co_mgr.getTVColor('data','default')
            # Default bar w/out color
            if horz:
                svg.append(f'<rect x="{x}" y="{y}" width="{bar_len}" height="{bar_sz}" fill="{_co}" />')
            else:
                svg.append(f'<rect x="{x}" y="{y-bar_len}" width="{bar_sz}" height="{bar_len}" fill="{_co}" />')
            # Colorize it
            if color_by is not None:
                quantities   = self.colorRenderOrder(df, color_by, count_by, count_by_set)
                value        = quantities.sum()['count'][0]
                quantities   = quantities.filter(pl.col('count') > value/bar_len)
                intersection = self.__myIntersection__(global_color_order['index'], quantities['index'])
                d = x if horz else y
                for cb_bin in intersection:
                    v = quantities.filter(pl.col('index') == cb_bin)['count'][0]
                    l = bar_len * v / value
                    if l >= 1.0:
                        _co = self.co_mgr.getColor(cb_bin)
                        if horz:
                            svg.append(f'<rect x="{d}" y="{y}" width="{l}" height="{bar_sz}" fill="{_co}" />')
                            d += l
                        else:
                            svg.append(f'<rect x="{x}" y="{d-l}" width="{bar_sz}" height="{l}" fill="{_co}" />')
                            d -= l
        return ''.join(svg)

    #
    # From https://www.geeksforgeeks.org/python-intersection-two-lists/
    #
    def __myIntersection__(self, lst1, lst2):
        temp = set(lst2)
        lst3 = [value for value in lst1 if value in temp]
        return lst3

    # Doesn't understand duplicates...
    #my_list_1 = [1, 2, 3, 10, 11, 12, 15, 18, 20, 20, 20]  # This ordering is kept
    #my_list_2 = [20, 0, 0, 0,  1, 11,  2,  7,  9, 100, 20]
    #intersection(my_list_1, my_list_2)

    #
    # svgText() - Render SVG Text In A Consistent Manner
    #
    def svgText(self,
                txt,
                x,
                y,
                txt_h,
                color    = None,
                anchor   = 'start',
                font     = None,
                rotation = None):
        if font  is None: font = self.default_font
        if color is None: color = self.co_mgr.getTVColor('label','defaultfg')
        txt = str(txt)

        _html_txt = html.escape(txt)
        # The following breaks JupyterLab in some configs ... 2024-01-04
        #if ' ' in _html_txt:
        #    _html_txt = _html_txt.replace(' ','&nbsp;')

        if rotation is not None:
            return f'<text x="{x}" text-anchor="{anchor}" y="{y}" font-family="{font}" fill="{color}" font-size="{txt_h}px"' + \
                   f' transform="rotate({rotation},{x},{y})">{_html_txt}</text>'
        else:
            return f'<text x="{x}" text-anchor="{anchor}" y="{y}" font-family="{font}" fill="{color}" font-size="{txt_h}px">{_html_txt}</text>'

    #
    # Empirically-derived font metrics -- see the next javascript code block on the initial generation of these numbers
    # ... this is really just a starting point
    # ... looks like correct for visual studio code with a txt_h of 19.5 // 2023-02-05
    # ... updated on 2023-06-28 to load them from a file...
    #
    _font_metrics_ = None

    #
    # Javascript used to generate the above... with a little bit of editing for the results to fit into a dictionary (copied from JS Console)...
    # ... used at https://jsfiddle.net
    #
    _font_metrics_js_ = """
let str = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
svg = "<svg width=\"256\" height=\"256\">"
for (i=0;i<str.length;i++) {
	 svg += "<text id=\"_test_" + str[i] + "\" x=\"50\" y=\"50\">" + str[i] + "</text>"
}
svg += "</svg>"
document.write(svg)

for (i=0;i<str.length;i++) {
	let elem = document.getElementById("_test_" + str[i]);
	let rect = elem.getBoundingClientRect();
	console.log("\'" + str[i] + "\':" + rect.width)
}
"""

    #
    # 2023-06-28 // Updated javascript that includes the first 4K unicode characters as well...
    # ... however, it doesn't fill in the space character... which is integer value 32
    #
    _font_metrics_unicode_js_ = """
function toHexFour(i) {
	my_s = i.toString(16)
  while (my_s.length < 4) { my_s = "0" + my_s; }
  return my_s
}
let sz = 16;
let s  = '';
for (i=32;i<4096;i++) {
	if (i==32) c = '&nbsp;';
  else       c = '&#x' + toHexFour(i) + ';';
	s = s + '<text id="txt' + i + '" x="5" y="32" font-family="ariel" font-size="'+sz+'px">' + c + '</text>';
}
document.write('<?xml version="1.0" encoding="UTF-8" standalone="no"?>' +
               '<svg width="64" height="64">' + s + '</svg>');
for (i=32;i<4096;i++) {
	let elem = document.getElementById("txt"+i);
	let rect = elem.getBoundingClientRect();
	console.log("" + i + "," + sz + "," + rect.width);
}
"""

    #
    # This version checks 4096 unicode characters (as above)... and then 
    # does all pairs from 32...128 // but it needs to be more... and even
    # with these limits, takes forever... lots of memory.
    #
    _font_metrics_unicode_kern_js_ = """
function toHexFour(i) {
	my_s = i.toString(16)
  while (my_s.length < 4) { my_s = "0" + my_s; }
  return my_s
}
let sz = 16;
let s  = '';
for (i=32;i<4096;i++) {
	if (i==32) c = '&nbsp;';
  else       c = '&#x' + toHexFour(i) + ';';
	s += '<text id="txt' + i + '" x="5" y="32" font-family="ariel" font-size="'+sz+'px">' + c + '</text>';
}

for (i=32;i<128;i++) {
	if (i==32) c = '&nbsp;';
  else       c = '&#x' + toHexFour(i) + ';';
	for (j=32;j<128;j++) {
		if (j==32) d = '&nbsp;';
  	else       d = '&#x' + toHexFour(j) + ';';
		s += '<text id="txt' + i + '_' + j + '" x="5" y="32" font-family="ariel" font-size="'+sz+'px">' + c + d + '</text>';
  }
}
document.write('<?xml version="1.0" encoding="UTF-8" standalone="no"?>' +
               '<svg width="64" height="64">' + s + '</svg>');

const char_to_w = new Map();
for (i=32;i<4096;i++) {
	let elem = document.getElementById("txt"+i);
	let rect = elem.getBoundingClientRect();
  char_to_w[i] = rect.width;
  console.log("" + i + "," + rect.width)
}

for (i=32;i<128;i++) {
	for (j=32;j<128;j++) {
  	let elem = document.getElementById("txt"+i+"_"+j);
	  let rect = elem.getBoundingClientRect();
    if (rect.width != (char_to_w[i] + char_to_w[j])) {
    	console.log("" + i + "," + j + "," + rect.width)
	  }	    
  }
}
"""

    #
    # cropText() - Based on the height of the font, shorten the string to fit into a specific width...
    # ... empirically derived values for letters / so unlikely to work exactly right if the font changes
    #
    def cropText(self, txt, txt_h, w):
        # If it fits, it ships
        if self.textLength(txt,txt_h) <= w:
            return txt

        # Otherwise... iterate until it doesn't fit
        i = 1
        while self.textLength(txt[:i],txt_h) < w:
            i += 1
        
        # Assumption is the the '...' doesn't add too much...
        if i == 0:
            i += 1
        return txt[:i-1] + '...'

    #
    # To adjust, see the "rt_test_fontmetrics_2.ipynb" file
    # ... not really correct either... apparently monospace isn't monospaced (or at least not in firefox)
    #
    font_w_slope     = 0.54961443           # 0.5496575 
    font_w_intercept = 0.027930354650640865 # 0.1928203837408624
    
    #
    # textLength() - calculate the expected text length
    #
    def __Monospace__textLength__(self, txt, txt_h):
        char_w = self.font_w_slope * txt_h + self.font_w_intercept
        return len(txt)*char_w

    #
    # textLength() - calculate the expected text length
    # ... was eventuall modified for "Ariel" ... and then back to Times Serif since Ariel wasn't working with SVGLib
    # ... but the kerning made the calculation difficult... any lower case letters following an "f" were a problem...
    #
    def textLength(self, txt, txt_h):
        ratio14 = 15.8
        # Load the pre-calculated font-metrics
        if self._font_metrics_ is None:
            _rt_dir   = os.path.dirname(os.path.abspath(__file__))
            # _filename = os.path.join(_rt_dir, "../config", "20230629_ariel_14.txt")
            _filename = os.path.join(_rt_dir, "../config", "20230628_times_serif_14.txt")
            with open(_filename) as file:
                lines = [line.rstrip() for line in file]
            self._font_metrics_ = {}
            for _line in lines:
                _parts = _line.split(',')
                _ord   = int(_parts[0])
                _sz    = int(_parts[1])
                _w     = float(_parts[2])
                if _sz not in self._font_metrics_.keys():
                    self._font_metrics_[_sz] = {}    
                self._font_metrics_[_sz][_ord] = _w
            self._font_metrics_[_sz][32] = 4.0
        # Calculate the width of the specified string
        w = 0
        for c in txt:
            if ord(c) in self._font_metrics_[14].keys():
                w += self._font_metrics_[14][ord(c)] * txt_h/ratio14
            else:
                w += 10 * txt_h/ratio14
        return w

    #
    # renderBoxPlotColumn() - render a single boxplot column (originally from the TemporalBarchart Implementation)
    #
    def renderBoxPlotColumn(self, style, k_df, cx, yT, group_by_max, group_by_min, bar_w, count_by, color_by, cap_swarm_at):
        if self.isPandas(k_df):
            return self.__renderBoxPlotColumn_pandas__(style, k_df, cx, yT, group_by_max, group_by_min, bar_w, count_by, color_by, cap_swarm_at)
        elif self.isPolars(k_df):
            return self.__renderBoxPlotColumn_polars__(style, k_df, cx, yT, group_by_max, group_by_min, bar_w, count_by, color_by, cap_swarm_at)
        else:
            raise Exception('renderBoxPlotColumn() - not a pandas or polars dataframe')

    def __renderBoxPlotColumn_pandas__(self, style, k_df, cx, yT, group_by_max, group_by_min, bar_w, count_by, color_by, cap_swarm_at):
        svg = ''
        if len(k_df) > 0:
            color = self.co_mgr.getTVColor('data','default') 

            # Just plot points if less than 5...
            if len(k_df) < 5:
                x_sz = 3
                for _value in k_df[count_by]:
                    sy = yT(_value)
                    svg += f'<line x1="{cx-x_sz}" y1="{sy-x_sz}" x2="{cx+x_sz}" y2="{sy+x_sz}" stroke="{color}" stroke-width="2" />'
                    svg += f'<line x1="{cx-x_sz}" y1="{sy+x_sz}" x2="{cx+x_sz}" y2="{sy-x_sz}" stroke="{color}" stroke-width="2" />'
            else:
                # Derived partially from: https://byjus.com/maths/box-plot/
                _med           = k_df[count_by].median()
                q3             = k_df[count_by].quantile(0.75)
                q1             = k_df[count_by].quantile(0.25)
                iqr            = q3-q1                           # difference between 1st and 3rd quartile
                q3_plus_15iqr  = q3 + 1.5*iqr
                q1_minus_15iqr = q1 - 1.5*iqr

                # for uniform distributions... non-normal distributions, the tops and bottoms can exceed the max and mins...
                upper_color,upper_is_max = color,False
                if q3_plus_15iqr > group_by_max:
                    q3_plus_15iqr = group_by_max
                    upper_color,upper_is_max   = self.co_mgr.getTVColor('label','error'),True
                lower_color,lower_is_min = color,False
                if q1_minus_15iqr < group_by_min:
                    q1_minus_15iqr = group_by_min
                    lower_color,lower_is_min   = self.co_mgr.getTVColor('label','error'),True

                svg += f'<line x1="{cx-bar_w/2}" y1="{yT(q3_plus_15iqr)}"     x2="{cx+bar_w/2}"     y2="{yT(q3_plus_15iqr)}"    stroke="{upper_color}" stroke-width="1.5" />'
                svg += f'<rect  x="{cx-bar_w/2}"  y="{yT(q3)}"             width="{bar_w}"      height="{yT(q1)-yT(q3)}"        stroke="{color}"       stroke-width="1"   fill-opacity="0.0" />'
                svg += f'<line x1="{cx-bar_w/2}" y1="{yT(q3)}"                x2="{cx+bar_w/2}"     y2="{yT(q3)}"               stroke="{color}"       stroke-width="1.5" />'
                svg += f'<line x1="{cx-bar_w/2}" y1="{yT(_med)}"              x2="{cx+bar_w/2}"     y2="{yT(_med)}"             stroke="{color}"       stroke-width="1.5" />'
                svg += f'<line x1="{cx-bar_w/2}" y1="{yT(q1)}"                x2="{cx+bar_w/2}"     y2="{yT(q1)}"               stroke="{color}"       stroke-width="1.5" />'
                svg += f'<line x1="{cx-bar_w/2}" y1="{yT(q1_minus_15iqr)}"    x2="{cx+bar_w/2}"     y2="{yT(q1_minus_15iqr)}"   stroke="{lower_color}" stroke-width="1.5" />'

                svg += f'<line x1="{cx}"          y1="{yT(q3)}"                x2="{cx}"            y2="{yT(q3_plus_15iqr)}"    stroke="{upper_color}" stroke-width="0.5" />'
                if upper_is_max:
                    svg += f'<line x1="{cx}"      y1="{yT(q3_plus_15iqr)}"     x2="{cx+5}"          y2="{yT(q3_plus_15iqr)+5}"  stroke="{upper_color}" stroke-width="0.5" />'
                    svg += f'<line x1="{cx}"      y1="{yT(q3_plus_15iqr)}"     x2="{cx-5}"          y2="{yT(q3_plus_15iqr)+5}"  stroke="{upper_color}" stroke-width="0.5" />'
                svg += f'<line x1="{cx}"          y1="{yT(q1)}"                x2="{cx}"            y2="{yT(q1_minus_15iqr)}"   stroke="{lower_color}" stroke-width="0.5" />'
                if lower_is_min:
                    svg += f'<line x1="{cx}"      y1="{yT(q1_minus_15iqr)}"    x2="{cx+5}"          y2="{yT(q1_minus_15iqr)-5}" stroke="{upper_color}" stroke-width="0.5" />'
                    svg += f'<line x1="{cx}"      y1="{yT(q1_minus_15iqr)}"    x2="{cx-5}"          y2="{yT(q1_minus_15iqr)-5}" stroke="{upper_color}" stroke-width="0.5" />'

                # Add marks for any items that are outliers
                _df = k_df[(k_df[count_by] > q3_plus_15iqr) | (k_df[count_by] < q1_minus_15iqr)]
                for v in _df[count_by]:
                    if v > q3_plus_15iqr:
                        svg += f'<circle cx="{cx}" cy="{yT(v)}" r="1.5" fill="{color}" />'
                    if v < q3_plus_15iqr:
                        svg += f'<circle cx="{cx}" cy="{yT(v)}" r="1.5" fill="{color}" />'

                # Add the swarm elements
                if style == 'boxplot_w_swarm':
                    # Provide cap... because this could take forever for large dataframes
                    if cap_swarm_at is not None and len(k_df) > cap_swarm_at:
                        _df = k_df.sample(cap_swarm_at)
                    else:
                        _df = k_df

                    if color_by is None:
                        for v in _df[count_by]:
                            sy    = yT(v)
                            mycx = cx + random.random() * bar_w/2 - bar_w/4
                            svg += f'<line x1="{mycx-1}" y1="{sy-1}" x2="{mycx+1}" y2="{sy+1}" stroke="{color}" stroke-width="0.5" />'
                            svg += f'<line x1="{mycx-1}" y1="{sy+1}" x2="{mycx+1}" y2="{sy-1}" stroke="{color}" stroke-width="0.5" />'
                    else:
                        for ksw,ksw_df in _df.groupby(color_by):
                            my_color = self.co_mgr.getColor(ksw)
                            for v in ksw_df[count_by]:
                                sy    = yT(v)
                                mycx = cx + random.random() * bar_w/2 - bar_w/4
                                svg += f'<line x1="{mycx-1}" y1="{sy-1}" x2="{mycx+1}" y2="{sy+1}" stroke="{my_color}" stroke-width="0.5" />'
                                svg += f'<line x1="{mycx-1}" y1="{sy+1}" x2="{mycx+1}" y2="{sy-1}" stroke="{my_color}" stroke-width="0.5" />'
        return svg

    def __renderBoxPlotColumn_polars__(self, style, k_df, cx, yT, group_by_max, group_by_min, bar_w, count_by, color_by, cap_swarm_at):
        svg = ''
        if len(k_df) > 0:
            color = self.co_mgr.getTVColor('data','default') 

            # Just plot points if less than 5...
            if len(k_df) < 5:
                x_sz = 3
                for _value in k_df[count_by]:
                    sy = yT(_value)
                    svg += f'<line x1="{cx-x_sz}" y1="{sy-x_sz}" x2="{cx+x_sz}" y2="{sy+x_sz}" stroke="{color}" stroke-width="2" />'
                    svg += f'<line x1="{cx-x_sz}" y1="{sy+x_sz}" x2="{cx+x_sz}" y2="{sy-x_sz}" stroke="{color}" stroke-width="2" />'
            else:
                # Derived partially from: https://byjus.com/maths/box-plot/
                _med           = k_df[count_by].median()
                q3             = k_df[count_by].quantile(0.75)
                q1             = k_df[count_by].quantile(0.25)
                iqr            = q3-q1                           # difference between 1st and 3rd quartile
                q3_plus_15iqr  = q3 + 1.5*iqr
                q1_minus_15iqr = q1 - 1.5*iqr

                # for uniform distributions... non-normal distributions, the tops and bottoms can exceed the max and mins...
                upper_color,upper_is_max = color,False
                if q3_plus_15iqr > group_by_max:
                    q3_plus_15iqr = group_by_max
                    upper_color,upper_is_max   = self.co_mgr.getTVColor('label','error'),True
                lower_color,lower_is_min = color,False
                if q1_minus_15iqr < group_by_min:
                    q1_minus_15iqr = group_by_min
                    lower_color,lower_is_min   = self.co_mgr.getTVColor('label','error'),True

                svg += f'<line x1="{cx-bar_w/2}" y1="{yT(q3_plus_15iqr)}"     x2="{cx+bar_w/2}"     y2="{yT(q3_plus_15iqr)}"    stroke="{upper_color}" stroke-width="1.5" />'
                svg += f'<rect  x="{cx-bar_w/2}"  y="{yT(q3)}"             width="{bar_w}"      height="{yT(q1)-yT(q3)}"        stroke="{color}"       stroke-width="1"   fill-opacity="0.0" />'
                svg += f'<line x1="{cx-bar_w/2}" y1="{yT(q3)}"                x2="{cx+bar_w/2}"     y2="{yT(q3)}"               stroke="{color}"       stroke-width="1.5" />'
                svg += f'<line x1="{cx-bar_w/2}" y1="{yT(_med)}"              x2="{cx+bar_w/2}"     y2="{yT(_med)}"             stroke="{color}"       stroke-width="1.5" />'
                svg += f'<line x1="{cx-bar_w/2}" y1="{yT(q1)}"                x2="{cx+bar_w/2}"     y2="{yT(q1)}"               stroke="{color}"       stroke-width="1.5" />'
                svg += f'<line x1="{cx-bar_w/2}" y1="{yT(q1_minus_15iqr)}"    x2="{cx+bar_w/2}"     y2="{yT(q1_minus_15iqr)}"   stroke="{lower_color}" stroke-width="1.5" />'

                svg += f'<line x1="{cx}"          y1="{yT(q3)}"                x2="{cx}"            y2="{yT(q3_plus_15iqr)}"    stroke="{upper_color}" stroke-width="0.5" />'
                if upper_is_max:
                    svg += f'<line x1="{cx}"      y1="{yT(q3_plus_15iqr)}"     x2="{cx+5}"          y2="{yT(q3_plus_15iqr)+5}"  stroke="{upper_color}" stroke-width="0.5" />'
                    svg += f'<line x1="{cx}"      y1="{yT(q3_plus_15iqr)}"     x2="{cx-5}"          y2="{yT(q3_plus_15iqr)+5}"  stroke="{upper_color}" stroke-width="0.5" />'
                svg += f'<line x1="{cx}"          y1="{yT(q1)}"                x2="{cx}"            y2="{yT(q1_minus_15iqr)}"   stroke="{lower_color}" stroke-width="0.5" />'
                if lower_is_min:
                    svg += f'<line x1="{cx}"      y1="{yT(q1_minus_15iqr)}"    x2="{cx+5}"          y2="{yT(q1_minus_15iqr)-5}" stroke="{upper_color}" stroke-width="0.5" />'
                    svg += f'<line x1="{cx}"      y1="{yT(q1_minus_15iqr)}"    x2="{cx-5}"          y2="{yT(q1_minus_15iqr)-5}" stroke="{upper_color}" stroke-width="0.5" />'

                # Add marks for any items that are outliers
                _df = k_df.filter(pl.col(count_by) > q3_plus_15iqr)
                for v in _df[count_by]:
                    svg += f'<circle cx="{cx}" cy="{yT(v)}" r="1.5" fill="{color}" />'
                _df = k_df.filter(pl.col(count_by) < q1_minus_15iqr)
                for v in _df[count_by]:
                    svg += f'<circle cx="{cx}" cy="{yT(v)}" r="1.5" fill="{color}" />'

                # Add the swarm elements
                if style == 'boxplot_w_swarm':
                    # Provide cap... because this could take forever for large dataframes
                    if cap_swarm_at is not None and len(k_df) > cap_swarm_at:
                        _df = k_df.sample(cap_swarm_at)
                    else:
                        _df = k_df

                    if color_by is None:
                        for v in _df[count_by]:
                            sy    = yT(v)
                            mycx = cx + random.random() * bar_w/2 - bar_w/4
                            svg += f'<line x1="{mycx-1}" y1="{sy-1}" x2="{mycx+1}" y2="{sy+1}" stroke="{color}" stroke-width="0.5" />'
                            svg += f'<line x1="{mycx-1}" y1="{sy+1}" x2="{mycx+1}" y2="{sy-1}" stroke="{color}" stroke-width="0.5" />'
                    else:
                        for ksw,ksw_df in _df.groupby(color_by):
                            my_color = self.co_mgr.getColor(ksw)
                            for v in ksw_df[count_by]:
                                sy    = yT(v)
                                mycx = cx + random.random() * bar_w/2 - bar_w/4
                                svg += f'<line x1="{mycx-1}" y1="{sy-1}" x2="{mycx+1}" y2="{sy+1}" stroke="{my_color}" stroke-width="0.5" />'
                                svg += f'<line x1="{mycx-1}" y1="{sy+1}" x2="{mycx+1}" y2="{sy-1}" stroke="{my_color}" stroke-width="0.5" />'
        return svg

    # ===========================================================================================================================================================

    #
    # Calculate the angled position string top and bottom position
    #
    def calculateAngledLabelTopAndBottomPosition(self, x, y, bar_w, txt_h, angle):
        frac_vert,frac_horz,bar_y = angle/90, (90-angle)/90, 0
        as_rad = pi*(angle+90)/180.0 # more than just radian conversion...
        horz_tpos  = (x+4,               y+4)       # top of string begin if the string were rendered horizontally
        horz_bpos  = (x+4,               y+4+txt_h) # bottom of string begin if the string were rendered horizontally
        vert_tpos  = (x+bar_w/2+txt_h/2, y+4)       # top of string begin if the string were rendered vertically
        vert_bpos  = (x+bar_w/2-txt_h/2, y+4)       # bottom of string begin if the string were rendered vertically
        angle_tpos = (vert_tpos[0]*frac_vert + horz_tpos[0]*frac_horz, vert_tpos[1]*frac_vert + horz_tpos[1]*frac_horz)
        angle_bpos = (angle_tpos[0] + cos(as_rad)*txt_h,               angle_tpos[1] + sin(as_rad)*txt_h)
        return angle_tpos,angle_bpos

    #
    # Does the specified angle cause the label to not overlap with the next label?
    # ... there's a close formed solution here... but it's beyond me :(
    # ... so many wasted cpu cycles... so many...
    #
    # ... see the rt_test_rotated_label prototype for testing/derivation
    #
    def doesAngleWorkForLabel(self, bar_w, txt_h, angle):
        if angle < 0 or angle >= 90:
            raise Exception(f'RACETrack.doesAngleWorkForLabel() - angle must be between [0,90) ... supplied angle = {angle}')

        # Position of label 0 and then label 1
        angle0_tpos,angle0_bpos = self.calculateAngledLabelTopAndBottomPosition(0,    0, bar_w, txt_h, angle)
        angle1_tpos,angle1_bpos = self.calculateAngledLabelTopAndBottomPosition(bar_w,0, bar_w, txt_h, angle)

        # Line from angle0_tpos in the direction of the angle...  is it underneath the angle1_bpos?
        m = sin(pi*angle/180)
        b = angle0_tpos[1] - m*angle0_tpos[0]
        return (m*angle1_bpos[0] + b) > angle1_bpos[1]

    #
    # Best angle for rotated label?
    #
    def bestAngleForRotatedLabels(self, bar_w, txt_h):
        for angle in range(0,90):
            if self.doesAngleWorkForLabel(bar_w, txt_h, angle):
                return angle
        return 90
    
    #
    # Determine if a string is an integer
    # ... shouldn't be used at scale
    # ... there's got to be a better way :( ... or some kind of builtin
    #
    def strIsInt(self, x):
        try:
            int(x)
            return True
        except:
            return False

    #
    # Determine if a string is a float
    # ... shouldn't be used at scale
    # ... there's got to be a better way :( ... or some kind of builtin
    #
    def strIsFloat(self, x):
        try:
            float(x)
            return True
        except:
            return False




# Copyright 2022 David Trimm
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pandas as pd
import polars as pl
import numpy as np
import random
import re

from shapely.geometry import Polygon

from math import sqrt

from rt_component import RTComponent

__name__ = 'rt_periodic_barchart_mixin'

#
# Periodic BarChart Mixin
# - Most of this code is lifted from the TemporalBarChart mixin
#
class RTPeriodicBarChartMixin(object):
    #
    # Init / Constructor for this mixin
    #
    def __periodic_barchart_mixin_init__(self):
        # Only periods allowed
        self.time_periods      = ['quarter', 'month', 'day', 'day_of_week', 'day_of_week_hour', 'hour', 'minute', 'second']
        
        # Number of bins in each period
        self.time_periods_bins = [4,         12,      31,             7,             7*24,               24,     60,       60]
        
        # Ordering of the periods
        self.time_periods_strs  = []
        for period_i in range(0,len(self.time_periods)):
            period = self.time_periods[period_i]
            if   period == 'quarter':
                self.time_periods_strs.append(['Q1',  'Q2',  'Q3',  'Q4'])
            elif period == 'month':
                self.time_periods_strs.append(['Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec'])
            elif period == 'day': # day of month
                my_arr = []
                for i in range(1,32):
                    my_arr.append(f'{i:02}')
                self.time_periods_strs.append(my_arr)
            elif period == 'day_of_week':
                self.time_periods_strs.append(['Sun','Mon','Tue','Wed','Thu','Fri','Sat'])
            elif period == 'day_of_week_hour':
                my_arr = []
                for dow in ['Sun','Mon','Tue','Wed','Thu','Fri','Sat']:
                    for hr in range(0,24):
                        my_arr.append(dow + f'-{hr:02}')
                self.time_periods_strs.append(my_arr)
            elif period == 'hour' or period == 'minute' or period == 'second':
                my_arr = []
                for i in range(0,self.time_periods_bins[period_i]):
                    my_arr.append(f'{i:02}')
                self.time_periods_strs.append(my_arr)
                
        # Double check the bin counts...
        for period_i in range(0,len(self.time_periods)):
            if self.time_periods_bins[period_i] != len(self.time_periods_strs[period_i]):
                raise Exception(f'time period {self.time_periods[period_i]} not set correctly')

    #
    # periodicBarChartPreferredSize()
    # - Return the preferred size
    #
    def periodicBarChartPreferredDimensions(self, time_period=None, min_bar_w=2, **kwargs):
        if time_period is not None and time_period in self.time_periods:
            i    = self.time_periods.index(time_period)
            bins = self.time_periods_bins[i]
            w    = bins * min_bar_w + 15 # give each bar 3 pixels... add a little for inserts and labeling
            h    = int(0.33*w)
            if h < 48:
                h = 48
                w = 3*h
            return (w,h)
        return (384,128)

    #
    # periodicBarChartMinimumSize()
    # - Return the minimum size
    #
    def periodicBarChartMinimumDimensions(self, time_period=None, min_bar_w=2, **kwargs):
        if time_period is not None and time_period in self.time_periods:
            i    = self.time_periods.index(time_period)
            bins = self.time_periods_bins[i]
            w    = bins * min_bar_w + 15 # give each bar 2 pixels... add a little for inserts and labeling
            h    = int(0.33*w)
            if h < 24:
                h = 24
                w = 3*h
            return (w,h)
        return (256,96)

    #
    # periodicBarChartSmallMultipleSize()
    #
    def periodicBarChartSmallMultipleDimensions(self, time_period=None, min_bar_w=2, **kwargs):
        return (32,24)

    #
    # periodicBarChartRequiredFields()
    # - Required fields for this configuration of the periodic barchart
    #
    def periodicBarChartRequiredFields(self, **kwargs):
        columns_set = set()
        self.identifyColumnsFromParameters('ts_field', kwargs, columns_set) # may need additional work if "None"
        self.identifyColumnsFromParameters('color_by', kwargs, columns_set)
        self.identifyColumnsFromParameters('count_by', kwargs, columns_set)
        return columns_set

    #
    # periodicBarChart
    # 
    def periodicBarChart(self,
                         df,                                    # dataframe to render
                         # ------------------------------------ # everything else is a default...
                         time_period           = 'day_of_week', # periodicity to render
                         ts_field              = None,          # timestamp field // needs to be a np.datetime64 column...                         
                         color_by              = None,          # just the default color or a string for a field
                         global_color_order    = None,          # color by ordering... if none (default), will be created and filled in...                         
                         count_by              = None,          # none means just count rows, otherwise, use a field to sum by
                         count_by_set          = False,         # count by using a set operation                         
                         widget_id             = None,          # naming the svg elements                         
                         # ------------------------------------ #                         
                         global_max            = None,          # maximum to use for the bar heights
                         global_min            = None,          # minimum (only used for the boxplot style(s))                         
                         just_calc_max         = False,         # forces return of the maximum for this render config
                                                                # ... which will then be used for the global max across bar charts
                         # ------------------------------------ #
                         style                 = 'barchart',    # 'barchart' or 'boxplot' or 'boxplot_w_swarm'
                         cap_swarm_at          = 200,           # cap the swarm plot at the specified number... if set to None, then no caps
                         # ------------------------------------ # small multiple options
                         sm_type               = None,          # should be the method name // similar to the smallMultiples method
                         sm_w                  = None,          # override the width of the small multiple
                         sm_h                  = None,          # override the height of the small multiple
                         sm_params             = {},            # dictionary of parameters for the small multiples
                         sm_x_axis_independent = True,          # Use independent axis for x (xy, temporal, and linkNode)
                         sm_y_axis_independent = True,          # Use independent axis for y (xy, temporal, periodic, pie)                         
                         # ------------------------------------ #                
                         track_state           = False,         # state tracking for interactive filtering                         
                         x_view                = 0,             # x offset for the view
                         y_view                = 0,             # y offset for the view
                         w                     = 512,           # width of the view
                         h                     = 128,           # height of the view
                         h_gap                 = 0,             # gap between bars.. should be a zero or a one...
                         min_bar_w             = 2,             # minimum bar width
                         txt_h                 = 14,            # text height for the labels
                         x_ins                 = 3,             # x insert (on both sides of the drawing)
                         y_ins                 = 3,
                         draw_labels           = True,          # draw labels flag
                         draw_border           = True,          # draw a border around the bar chart
                         draw_context          = True):         # draw background hints about the years, months, days, etc.
        _params_ = locals().copy()
        _params_.pop('self')
        return self.RTPeriodicBarChart(self, **_params_)

    #
    # RTPeriodicBarChart()
    #
    class RTPeriodicBarChart(RTComponent):
        #
        # Constructor
        #
        def __init__(self,
                     rt_self,
                     **kwargs):
            self.parms     = locals().copy()
            self.rt_self   = rt_self
            self.df        = rt_self.copyDataFrame(kwargs['df'])
            self.widget_id = kwargs['widget_id']
            
            # Make a widget_id if it's not set already
            if self.widget_id is None:
                self.widget_id = "periodicbarchart_" + str(random.randint(0,65535))

            self.time_period           = kwargs['time_period']
            self.ts_field              = kwargs['ts_field']
            self.color_by              = kwargs['color_by']
            self.global_color_order    = kwargs['global_color_order']
            self.count_by              = kwargs['count_by']
            self.count_by_set          = kwargs['count_by_set']
            self.global_max            = kwargs['global_max']
            self.global_min            = kwargs['global_min']
            self.style                 = kwargs['style']
            self.cap_swarm_at          = kwargs['cap_swarm_at']
            self.sm_type               = kwargs['sm_type']
            self.sm_w                  = kwargs['sm_w']
            self.sm_h                  = kwargs['sm_h']
            self.sm_params             = kwargs['sm_params']
            self.sm_x_axis_independent = kwargs['sm_x_axis_independent']
            self.sm_y_axis_independent = kwargs['sm_y_axis_independent']
            self.track_state           = kwargs['track_state']
            self.x_view                = kwargs['x_view']
            self.y_view                = kwargs['y_view']
            self.w                     = kwargs['w']
            self.h                     = kwargs['h']
            self.h_gap                 = kwargs['h_gap']
            self.min_bar_w             = kwargs['min_bar_w']
            self.txt_h                 = kwargs['txt_h']
            self.x_ins                 = kwargs['x_ins']
            self.y_ins                 = kwargs['y_ins']
            self.draw_labels           = kwargs['draw_labels']
            self.draw_border           = kwargs['draw_border']
            self.draw_context          = kwargs['draw_context']

            # Determine the timestamp field
            if self.ts_field is None:
                self.ts_field = rt_self.guessTimestampField(self.df)
            
            # Determine the periodicity index
            self.time_period = kwargs['time_period']
            self.period_i = rt_self.time_periods.index(kwargs['time_period'])

            # Perform the transforms
            # Apply count-by transofmrs
            if self.count_by is not None and rt_self.isTField(self.count_by):
                self.df,self.count_by = rt_self.applyTransform(self.df, self.count_by)

            # Apply color-by transforms
            if self.color_by is not None and rt_self.isTField(self.color_by):
                self.df,self.color_by = rt_self.applyTransform(self.df, self.color_by)

            # Check the count_by column
            if self.count_by_set == False:
                self.count_by_set = rt_self.countBySet(self.df, self.count_by)
            
            # Geometry lookup for tracking state
            self.geom_to_df = {}
            self.last_render = None

        #
        # dataFrameRanges() - calculate the max and min of the bars
        #
        def dataFrameRanges(self, period_field):
            if   self.rt_self.isPandas(self.df):
                return self.__dataFrameRanges_pandas__(period_field)
            elif self.rt_self.isPolars(self.df):
                return self.__dataFrameRanges_polars__(period_field)
            else:
                raise Exception('RTPeriodicBarChart.dataFrameRanges() -- only pandas and polars supported')

        #
        def __dataFrameRanges_pandas__(self, period_field):
            group_by = self.df.groupby(period_field)
            if   self.count_by is None:
                group_by_min = 0
                group_by_max = group_by.size().max()
            elif self.count_by_set:
                _df         = self.df.groupby([period_field,self.count_by]).size().reset_index()
                _df_for_max = _df.groupby(period_field)
                group_by_min = 0
                group_by_max = _df_for_max.size().max()
            elif self.style.startswith('boxplot'):
                group_by_min = self.df[self.count_by].min()                    
                group_by_max = self.df[self.count_by].max()
            else:
                group_by_min = 0
                group_by_max = group_by[self.count_by].sum().max()
            return group_by, group_by_min, group_by_max

        #
        def __dataFrameRanges_polars__(self, period_field):
            group_by_min = 0
            if   self.count_by is None:
                tmp_df = self.df.drop(set(self.df.columns) - set([period_field])) \
                                .group_by(period_field)                           \
                                .agg(pl.count()                                   \
                                       .alias('__count__'))
                group_by_max = tmp_df['__count__'].max()
            elif self.count_by_set:
                tmp_df = self.df.drop(set(self.df.columns) - set([period_field, self.count_by])) \
                                .group_by(period_field)                                          \
                                .n_unique()
                group_by_max = tmp_df[self.count_by].max()
            elif self.style.startswith('boxplot'):
                group_by_min = self.df[self.count_by].min()                    
                group_by_max = self.df[self.count_by].max()
            else:
                tmp_df = self.df.drop(set(self.df.columns) - set([period_field, self.count_by])) \
                                .group_by(period_field)                                          \
                                .agg(pl.sum(self.count_by)                                       \
                                       .alias('__count__'))
                group_by_max = tmp_df['__count__'].max()
            return self.df.group_by(period_field), group_by_min, group_by_max

        #
        # print() version of class
        #
        def __repr__(self):
            def tQontQ(t): return 'None' if t is None else "'" + str(t) + "'"
            return f'periodicBarChart(df.len={len(self.df)}, ts_field=\'{self.ts_field}\', period=\'{self.time_period}\', count_by={tQontQ(self.count_by)}, color_by={tQontQ(self.color_by)}, {self.w}x{self.h})'

        #
        # SVG Representation Renderer
        #
        def _repr_svg_(self):
            if self.last_render is None: self.renderSVG()
            return self.last_render

        #
        # renderSVG() - create the SVG
        #
        def renderSVG(self, just_calc_max=False):
            if self.track_state:
                self.geom_to_df = {}

            # Color ordering
            if self.global_color_order is None:
                self.global_color_order = self.rt_self.colorRenderOrder(self.df, self.color_by, self.count_by, self.count_by_set)

            # If the height/width are less than the minimums, turn off labeling... and make the min_bar_w = 1
            # ... for small multiples
            params_orig_minus_self = self.parms.copy()
            params_orig_minus_self.pop('self')
            min_dims = self.rt_self.periodicBarChartSmallMultipleDimensions(**params_orig_minus_self)
            if self.w <= min_dims[0] or self.h <= min_dims[1]:
                self.draw_labels = False
                self.min_bar_w   = 1
                self.x_ins       = 1
                self.y_ins       = 1
                self.h_gap       = 0

            # Limited range of pixels are allowed...
            if   self.h_gap > 4:
                self.h_gap = 4
            elif self.h_gap < 0:
                self.h_gap = 0
            
            # Calculate the usable width
            w_usable = self.w - 2*self.x_ins
            x_left   = self.x_ins
            if self.draw_labels:
                x_left    =           2*self.y_ins + self.txt_h
                w_usable  = self.w - (3*self.y_ins + self.txt_h)

            # Create a new field with the periodic value there
            t_field = self.rt_self.createTField(self.ts_field, self.time_period)
            self.df, period_field = self.rt_self.applyTransform(self.df, t_field)

            # Total number of bins
            bins    = self.rt_self.time_periods_bins[self.period_i]

            # Adjust the min_bar_w if this is small multiples are to be included
            if self.sm_type is not None:
                if self.sm_w is None or self.sm_h is None:
                    self.sm_w,self.sm_h = getattr(self.rt_self, f'{self.sm_type}SmallMultipleDimensions')(**self.sm_params)
                min_bar_w = self.sm_w # do we even use min_bar_w anywhere?
                bar_w   = ((w_usable - (bins*self.h_gap))/bins)
                if   bar_w <  self.sm_w:
                    bar_w = self.sm_w                
                    if self.draw_labels:
                        self.w   =           3*self.y_ins + self.txt_h + bins*bar_w
                        w_usable = self.w - (3*self.y_ins + self.txt_h)
                    else:
                        w        =           2*self.y_ins              + bins*bar_w
                        w_usable = self.w - (2*self.y_ins)
                elif bar_w == self.sm_w:
                    pass
                else:
                    sm_prop   = self.sm_h/self.sm_w
                    self.sm_w = bar_w
                    self.sm_h = bar_w * sm_prop
            else:
                # Finalize the bar width
                bar_w   = ((w_usable - (bins*self.h_gap))/bins)

            # Height geometry
            sm_cy = 0
            if self.sm_type is not None:
                max_bar_h = self.h - 2*self.y_ins             - self.sm_h
                sm_cy     = self.h -   self.y_ins - max_bar_h - self.sm_h/2
                if self.draw_labels:
                    max_bar_h = self.h - 2*self.y_ins - self.sm_h - self.txt_h - 2
            else:
                max_bar_h = self.h - 2*self.y_ins
                if self.draw_labels:
                    max_bar_h = self.h - 2*self.y_ins - self.txt_h - 2

            # Determine the max
            group_by, group_by_min, group_by_max = self.dataFrameRanges(period_field)
            if self.global_max is not None:
                group_by_min, group_by_max = self.global_min, self.global_max
            if just_calc_max:
                return group_by_min,group_by_max

            # Start the SVG Frame
            svg = f'<svg id="{self.widget_id}" x="{self.x_view}" y="{self.y_view}" width="{self.w}" height="{self.h}" xmlns="http://www.w3.org/2000/svg">'
            background_color = self.rt_self.co_mgr.getTVColor('background','default')
            svg += f'<rect width="{self.w-1}" height="{self.h-1}" x="0" y="0" fill="{background_color}" stroke="{background_color}" />'
            
            # Draw the background for the temporal chart
            textfg = self.rt_self.co_mgr.getTVColor('label','defaultfg')
            y_baseline = self.h - self.y_ins - 1
            if self.draw_labels:
                y_baseline = self.h - self.y_ins - self.txt_h - 1

            # Draw temporal context
            if self.draw_context:
                pass
            
            sm_adj = 0
            if self.sm_type is not None:
                sm_adj = self.sm_h

            # Draw the axes
            axis_co = self.rt_self.co_mgr.getTVColor('axis','default')
            svg += f'<line x1="{x_left}" y1="{y_baseline+1}" x2="{x_left}"            y2="{self.y_ins+sm_adj}" stroke="{axis_co}" stroke-width="1" />'
            svg += f'<line x1="{x_left}" y1="{y_baseline+1}" x2="{x_left + w_usable}" y2="{y_baseline+1}"      stroke="{axis_co}" stroke-width="1" />'

            # Draw the bars
            if self.style == 'barchart':
                for k,k_df in group_by:
                    if type(k) == tuple and len(k) == 1: k = k[0] # polars fixes on 2024-07-19
                    x = x_left + 1 + (bar_w+self.h_gap)*self.rt_self.time_periods_strs[self.period_i].index(k)
                    if   self.count_by is None:
                        px = max_bar_h * len(k_df) / group_by_max
                    elif self.count_by_set:
                        px = max_bar_h * len(set(k_df[self.count_by])) / group_by_max
                    else:
                        px = max_bar_h * k_df[self.count_by].sum() / group_by_max
                            
                    svg += self.rt_self.colorizeBar(k_df, self.global_color_order, self.color_by, self.count_by, self.count_by_set, x, y_baseline, px, bar_w, False)

                    if self.track_state:
                        _poly = Polygon([[x,y_baseline],[x+bar_w,y_baseline],[x+bar_w,y_baseline-px],[x,y_baseline-px]])
                        self.geom_to_df[_poly] = k_df

            elif self.style.startswith('boxplot'):
                # Adjust the bar width to something reasonable
                _bar_w = bar_w
                if _bar_w > 16:
                    _bar_w = 16
                
                # Make a y-transform lambda expression
                if group_by_max == 0:
                    group_by_max = 1
                yT = lambda __y__: (y_baseline - max_bar_h * (__y__ - group_by_min) / (group_by_max - group_by_min))

                # Render the boxplot columns
                for k,k_df in group_by:
                    x = x_left + 1 + (bar_w+self.h_gap)*self.rt_self.time_periods_strs[self.period_i].index(k)
                    _cx = x + bar_w/2

                    svg += self.rt_self.renderBoxPlotColumn(self.style, k_df, _cx, yT, group_by_max, group_by_min, _bar_w, self.count_by, self.color_by, self.cap_swarm_at)

                    if self.track_state:
                        _poly = Polygon([[x,y_baseline],[x+bar_w,y_baseline],[x+bar_w,y_baseline-max_bar_h],[x,y_baseline-max_bar_h]])
                        self.geom_to_df[_poly] = k_df

            else:
                raise Exception(f'RTPeriodicBarChart() - unknown style "{self.style}"')


            # Handle the small multiple renders
            if self.sm_type is not None:
                if   self.rt_self.isPandas(self.df): group_by = self.df.groupby(period_field)
                elif self.rt_self.isPolars(self.df): group_by = self.df.group_by(period_field)

                node_to_xy  = {}
                node_to_dfs = {}

                for key,key_df in group_by:
                    if type(key) == tuple and len(key) == 1: key = key[0] # polars fixes on 2024-07-19
                    x = x_left + 1 + (bar_w+self.h_gap)*self.rt_self.time_periods_strs[self.period_i].index(key)
                    if len(key_df) != 0:
                        node_to_xy[key]  = [x + bar_w/2, sm_cy]
                        node_to_dfs[key] = key_df
                
                sm_lu = self.rt_self.createSmallMultiples(self.df, node_to_dfs, node_to_xy,
                                                          self.count_by, self.count_by_set, self.color_by, self.ts_field, self.widget_id,
                                                          self.sm_type, self.sm_params, self.sm_x_axis_independent, self.sm_y_axis_independent,
                                                          self.sm_w, self.sm_h)

                for node_str in sm_lu.keys():
                    svg += sm_lu[node_str]

            # Draw the labels // mirrors the rt_temporal_barchart_mixin codeblock
            if self.draw_labels:
                _label_x_min_  = self.rt_self.time_periods_strs[self.period_i][0]
                _label_x_max_  = self.rt_self.time_periods_strs[self.period_i][-1]
                _label_x_axis_ = self.rt_self.time_periods[self.period_i]
                _len_x_min_, _len_x_max_ = self.rt_self.textLength(_label_x_min_,  self.txt_h), self.rt_self.textLength(_label_x_max_, self.txt_h)
                _len_x_axis_             = self.rt_self.textLength(_label_x_axis_, self.txt_h)

                if (_len_x_min_ + _len_x_max_ + _len_x_axis_ + 20) < w_usable:
                    svg += self.rt_self.svgText(_label_x_min_,   x_left,              self.h-3, self.txt_h)
                    svg += self.rt_self.svgText(_label_x_max_,   self.w - self.x_ins, self.h-3, self.txt_h, anchor='end')
                    svg += self.rt_self.svgText(_label_x_axis_,  x_left + w_usable/2, self.h-3, self.txt_h, anchor='middle')
                elif (_len_x_axis_ + 10) < w_usable:
                    svg += self.rt_self.svgText(_label_x_axis_,  x_left + w_usable/2, self.h-3, self.txt_h, anchor='middle')

                # Max Label
                _str_max,_str_min = f'{group_by_max:{self.rt_self.fformat}}',''
                if re.match('.*\.0*',_str_max):
                    _str_max = _str_max[:_str_max.index('.')]
                svg += self.rt_self.svgText(_str_max,     self.x_ins+self.txt_h, self.y_ins + sm_adj,             self.txt_h, anchor='end',    rotation=-90)

                # Min Label (for boxplot only)
                if self.style.startswith('boxplot'):
                    _str_min = f'{group_by_min:{self.rt_self.fformat}}'
                    if re.match('.*\.0*',_str_min):
                        _str_min = _str_min[:_str_min.index('.')]
                    svg += self.rt_self.svgText(_str_min, self.x_ins+self.txt_h, self.y_ins + sm_adj + max_bar_h, self.txt_h, anchor='start',  rotation=-90)

                # Count By Label
                _count_by_str = 'rows'
                if self.count_by:
                    _count_by_str = self.count_by
                if self.rt_self.textLength(_str_max, self.txt_h) + self.rt_self.textLength(_str_min, self.txt_h) + self.rt_self.textLength(_count_by_str, self.txt_h) + 10 < max_bar_h:
                    if self.style.startswith('boxplot'):
                        _mid_y = self.rt_self.textLength(_str_max, self.txt_h) + (max_bar_h - self.rt_self.textLength(_str_max, self.txt_h) - self.rt_self.textLength(_str_min, self.txt_h))/2
                        svg += self.rt_self.svgText(_count_by_str, self.x_ins+self.txt_h, 
                                                    self.y_ins + sm_adj + _mid_y, 
                                                    self.txt_h, anchor='middle',  rotation=-90)
                    else:
                        svg += self.rt_self.svgText(_count_by_str, self.x_ins+self.txt_h, self.y_ins + sm_adj + max_bar_h,   self.txt_h, anchor='start',  rotation=-90)

            # Draw the border
            if self.draw_border:
                border_color = self.rt_self.co_mgr.getTVColor('border','default')
                svg += f'<rect width="{self.w-1}" height="{self.h-1}" x="0" y="0" fill-opacity="0.0" stroke="{border_color}" />'

            svg += '</svg>'
            self.last_render = svg
            return svg

        #
        # smallMultipleFeatureVector()
        # ... feature vector for comparison with other small multiple instances of this class
        # ... pretty much a copy of the render code above...
        #
        def smallMultipleFeatureVector(self):
            # Create a new field with the periodic value there
            t_field = self.rt_self.createTField(self.ts_field, self.time_period)
            self.df, period_field = self.rt_self.applyTransform(self.df, t_field)

            # Total number of bins
            bins    = self.rt_self.time_periods_bins[self.period_i]

            # Determine the max
            group_by, group_by_min, group_by_max = self.dataFrameRanges(period_field)
            if self.global_max is not None:
                self.global_min, self.global_max = group_by_min, group_by_max
            
            # Iterate over the keys
            max_bar_h,fv = 1.0,{}
            for k,k_df in group_by:
                if   self.count_by is None:
                    px = max_bar_h * len(k_df) / group_by_max
                elif self.count_by_set:
                    px = max_bar_h * len(set(k_df[self.count_by])) / group_by_max
                else:
                    px = max_bar_h * k_df[self.count_by].sum() / group_by_max
                fv[k] = px
            
            # Make it into a unit vector
            sq_sum = 0
            for k in fv.keys():
                sq_sum += fv[k]*fv[k]
            sq_sum = sqrt(sq_sum)
            if sq_sum < 0.001:
                sq_sum = 0.001
            fv_norm = {}
            for k in fv.keys():
                fv_norm[k] = fv[k]/sq_sum

            return fv_norm

        #
        # Determine which dataframe geometries overlap with a specific
        #
        def overlappingDataFrames(self, to_intersect):
            _dfs = []
            for _poly in self.geom_to_df.keys():
                if _poly.intersects(to_intersect):
                    _dfs.append(self.geom_to_df[_poly])
            if len(_dfs) > 0:
                return self.rt_self.concatDataFrames(_dfs)
            else:
                return None



# Copyright 2024 David Trimm
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pandas as pd
import polars as pl
import numpy as np
import json

__name__ = 'rt_json_mixin'

#
# JSON Utilities Mixin
#
class RTJSONMixin(object):
    #
    # jsonRepr() - create an RTJSON object from a JSON reference
    #
    def jsonRepr(self, json_ref):
        if type(json_ref) == str:  json_ref = json.loads(json_ref)
        return self.RTJSON(self, json_ref)

    #
    # RTJSON() - class to parse JSON reference
    #
    class RTJSON(object):
        #
        # __init__()
        #
        def __init__(self, rt_self, _json_):
            self.rt_self    = rt_self
            self.fmto_stars = {}
            self.fmto_abs   = {}
            self.absEOL     = {}
            self.starEOL    = {}
            if type(_json_) == list:
                for i in range(len(_json_)):
                    self.__recursiveParseJSON__('$', '$', _json_[i])
            else:
                self.__recursiveParseJSON__('$', '$', _json_)

        #
        # __updateFMTO__()
        #
        def __updateFMTO__(self, _parent_, _child_, _abs_):
            if _abs_:
                if _parent_ not in self.fmto_abs: self.fmto_abs[_parent_] = []
                self.fmto_abs[_parent_].append(_child_)
            else:
                if _parent_ not in self.fmto_stars: self.fmto_stars[_parent_] = set()
                self.fmto_stars[_parent_].add(_child_)

        #
        # __updateEOL__() - update end of the line (actual leaf values of the JSON tree structure)
        #
        def __updateEOL__(self, _path_abs_, _path_star_, _value_):
            self.absEOL[_path_abs_] = _value_
            if _path_star_ not in self.starEOL: self.starEOL[_path_star_] = []
            self.starEOL[_path_star_].append(_value_)

        #
        # __recursiveParseJSON__()
        #
        def __recursiveParseJSON__(self, _path_abs_, _path_star_, _json_):
            if type(_json_) == dict:
                for k in _json_.keys():
                    _path_abs_child_, _path_star_child_  = _path_abs_+'.'+k, _path_star_+'.'+k
                    self.__updateFMTO__(_path_star_, _path_star_child_, False), self.__updateFMTO__(_path_abs_, _path_abs_child_, True)
                    self.__recursiveParseJSON__(_path_abs_child_, _path_star_child_, _json_[k])
            elif type(_json_) == list:
                _path_star_child_ = _path_star_+'[*]'
                self.__updateFMTO__(_path_star_, _path_star_child_, False)
                for i in range(len(_json_)):
                    _path_abs_child_  = _path_abs_+'['+str(i)+']'
                    self.__updateFMTO__(_path_abs_, _path_abs_child_, True)
                    self.__recursiveParseJSON__(_path_abs_child_, _path_star_child_, _json_[i])
            elif type(_json_) == str:
                self.__updateFMTO__(_path_star_, _path_star_+'-<str>', False), self.__updateFMTO__(_path_abs_, _path_abs_+f'-<str> {_json_}', True)
                self.__updateEOL__(_path_abs_, _path_star_, _json_)
            elif type(_json_) == int:
                self.__updateFMTO__(_path_star_, _path_star_+'-<int>', False), self.__updateFMTO__(_path_abs_, _path_abs_+f'-<int> {_json_}', True)
                self.__updateEOL__(_path_abs_, _path_star_, _json_)
            elif type(_json_) == float:
                self.__updateFMTO__(_path_star_, _path_star_+'-<float>', False), self.__updateFMTO__(_path_abs_, _path_abs_+f'-<float> {_json_}', True)
                self.__updateEOL__(_path_abs_, _path_star_, _json_)
            elif type(_json_) == bool:
                self.__updateFMTO__(_path_star_, _path_star_+'-<bool>', False), self.__updateFMTO__(_path_abs_, _path_abs_+f'-<bool> {_json_}', True)
                self.__updateEOL__(_path_abs_, _path_star_, _json_)
            elif _json_ is None:
                self.__updateFMTO__(_path_star_, _path_star_+'-[NONE]', False), self.__updateFMTO__(_path_abs_, _path_abs_+'-[NONE]', True)
                self.__updateEOL__(_path_abs_, _path_star_, _json_)
            else:
                raise Exception(f'Unknown type ("{type(_json_)}") for ("{_json_}") encountered in parseJSON()')

        #
        # __nodeLabels__()
        #
        def __nodeLabels__(self, _df_):
            _labels_ = {}
            _nodes_  = set(_df_['fm']) | set(_df_['to'])
            for _node_ in _nodes_:
                if   _node_.endswith('[*]'):    _labels_[_node_] = '[*]'
                elif _node_.endswith('<str>')   or \
                    _node_.endswith('<int>')    or \
                    _node_.endswith('<float>')  or \
                    _node_.endswith('<bool>'): _labels_[_node_]  = _node_[_node_.rindex('<'):]
                elif _node_.endswith('[NONE]'): _labels_[_node_] = 'None::None'
                elif '-<str> '   in _node_:     _labels_[_node_] = _node_[_node_.rindex('-<str> ')   +7:]
                elif '-<int> '   in _node_:     _labels_[_node_] = _node_[_node_.rindex('-<int> ')   +7:]
                elif '-<float> ' in _node_:     _labels_[_node_] = _node_[_node_.rindex('-<float> ') +9:]
                elif '-<bool> '  in _node_:     _labels_[_node_] = _node_[_node_.rindex('-<bool> ')  +8:]
                elif '.'         in _node_:     _labels_[_node_] = _node_[_node_.rindex('.')+1:]
                elif _node_ == '$':             _labels_[_node_] = 'ROOT'
            return _labels_
        #
        # absolutePathGraphDataFrame()
        #
        def absolutePathGraphDataFrame(self):
            _df_     = self.rt_self.graphDictToDataFrame(self.fmto_abs)
            _labels_ = self.__nodeLabels__(_df_)
            return _df_, [('fm','to')], _labels_

        #
        # starGraphDataFrame()
        #
        def starPathGraphDataFrame(self):
            _df_     = self.rt_self.graphDictToDataFrame(self.fmto_stars)
            _labels_ = self.__nodeLabels__(_df_)
            return _df_, [('fm','to')], _labels_




from setuptools import setup, find_packages
setup(
name='rtsvg',
version='0.1.15',
author='D. Trimm',
author_email='dave.trimm@gmail.com',
description='Visualization Package Leveraging SVG in Jupyter Notebooks',
packages=['rtsvg'],
classifiers=[
'Programming Language :: Python :: 3',
'License :: OSI Approved :: Apache Software License',
'Operating System :: OS Independent',
],
python_requires='>=3.11',
url='https://github.com/datrcode/racetrack_svg_framework'
)




# Copyright 2022 David Trimm
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re
import inspect
import random

import pandas as pd
import polars as pl
import numpy as np

from shapely.affinity import translate
from shapely.geometry import Polygon, Point

from rt_component import RTComponent

from math import ceil

__name__ = 'rt_layouts_mixin'

#
# Abstraction for colorscale // by default this is the dark theme
#
class RTLayoutsMixin(object):
    #
    # Place a node within the tree based on the layout path
    #
    def __recursivePlace(self, root, pos, leaf):
        if '|' in pos:
            first = pos[:pos.index('|')]
            rest  = pos[pos.index('|')+1:]
            if first not in root.keys():
                root[first] = {}
            self.__recursivePlace(root[first], rest, leaf)
        else:
            if pos in root.keys():
                print('error ... pos in {root.keys()}')
            else:
                root[pos] = leaf

    #
    # Return the layout type
    #
    def __layoutType(self, x):
        if x in ['c','n','e','w','s']: # BorderLayout
            return 'borderLayout'
        elif ',' in x: # GridLayout
            return 'gridLayout'
        elif re.match(r"[-]{0,1}[0-9]+",x) and int(x) < 0: # FlowLayout Horizontal
            return 'flowLayoutHorizontal'
        elif re.match(r"[-]{0,1}[0-9]+",x): # FlowLayout Vertical
            return 'flowLayoutVertical'
        else: # Unknown -- throw exception
            raise Exception(f'unknown layout mnemonic "{x}"')

    #
    # Flow Layout (Vertical) Calculation
    #
    def __flowLayoutVerticalCalculation(self, my_node, leaf_dims, so_far):
        min_dim  = None
        pref_dim = None
        for key in my_node.keys():
            my_min_dim  = leaf_dims[so_far + '|' + key][0]
            my_pref_dim = leaf_dims[so_far + '|' + key][1]
            if min_dim is None:
                min_dim  = my_min_dim
                pref_dim = my_pref_dim
            else:
                min_dim  = (max(min_dim[0],  my_min_dim[0]),  min_dim[1]  + my_min_dim[1])
                pref_dim = (max(pref_dim[0], my_pref_dim[0]), pref_dim[1] + my_pref_dim[1])
        return (min_dim, pref_dim)

    #
    # Flow Layout (Horizontal) Calculation
    #
    def __flowLayoutHorizontalCalculation(self, my_node, leaf_dims, so_far):
        min_dim  = None
        pref_dim = None
        for key in my_node.keys():
            my_min_dim  = leaf_dims[so_far + '|' + key][0]
            my_pref_dim = leaf_dims[so_far + '|' + key][1]
            if min_dim is None:
                min_dim  = my_min_dim
                pref_dim = my_pref_dim
            else:
                min_dim  = (min_dim[0]  + my_min_dim[0],  max(min_dim[1],  my_min_dim[1]))
                pref_dim = (pref_dim[0] + my_pref_dim[0], max(pref_dim[1], my_pref_dim[1]))
        return (min_dim, pref_dim)

    #
    # Grid Layout Calculation
    #
    def __gridLayoutCalculation(self, my_node, leaf_dims, so_far):
        min_dim  = None
        pref_dim = None
        max_x    = 1
        max_y    = 1
        for key in my_node.keys():
            x = int(key.split(',')[0])
            y = int(key.split(',')[1])
            max_x = max(x, max_x)
            max_y = max(y, max_y)

            my_min_dim  = leaf_dims[so_far + '|' + key][0]
            my_pref_dim = leaf_dims[so_far + '|' + key][1]
            if min_dim is None:
                min_dim  = my_min_dim
                pref_dim = my_pref_dim
            else:
                min_dim  = (max(min_dim[0],  my_min_dim[0]),  max(min_dim[1],  my_min_dim[1]))
                pref_dim = (max(pref_dim[0], my_pref_dim[0]), max(pref_dim[1], my_pref_dim[1]))


        return ((min_dim[0] * max_x, min_dim[1] * max_y), (pref_dim[0] * max_x, pref_dim[1] * max_y))

    #
    # Border Layout Calculation
    #
    def __borderLayoutCalculation(self, my_node, leaf_dims, so_far):
        s_dim = n_dim = w_dim = e_dim = c_dim = ((0,0),(0,0))
        if 'e' in my_node.keys():
            e_dim = leaf_dims[so_far+'|'+'e']
        if 'w' in my_node.keys():
            w_dim = leaf_dims[so_far+'|'+'w']
        if 'n' in my_node.keys():
            n_dim = leaf_dims[so_far+'|'+'n']
        if 's' in my_node.keys():
            s_dim = leaf_dims[so_far+'|'+'s']
        if 'c' in my_node.keys():
            c_dim = leaf_dims[so_far+'|'+'c']

        x_min  = max(e_dim[0][0] + c_dim[0][0] + w_dim[0][0], n_dim[0][0], s_dim[0][0])
        x_pref = max(e_dim[1][0] + c_dim[1][0] + w_dim[1][0], n_dim[1][0], s_dim[1][0])

        y_min  = max(e_dim[0][1],  c_dim[0][1]  ,w_dim[0][1]) + n_dim[0][1] + s_dim[0][1]
        y_pref = max(e_dim[1][1],  c_dim[1][1]  ,w_dim[1][1]) + n_dim[1][1] + s_dim[1][1]

        return ((x_min,y_min),(x_pref,y_pref))

    #
    # Fill in the interior node layout dimensions
    #
    def __fillInteriorNodeDimensions(self, my_node, leaf_dims, dims, so_far='root'):        
        #
        # Leaf Node -- store the dimensions
        #
        if type(my_node) == tuple:
            widget_type       = str(my_node)
            leaf_dims[so_far] = (dims[widget_type]['min'],dims[widget_type]['pref'])

        #
        # Leaf Node -- that doesn't really have a tuple...
        #
        elif type(my_node) == str:
            widget_type       = my_node
            leaf_dims[so_far] = (dims[widget_type]['min'],dims[widget_type]['pref'])
            
        #
        # Interior Node -- have the leaves execute first... then do the layout calculation here
        #
        else:
            # Visit leaves first... and then process this node
            for key in my_node.keys():
                self.__fillInteriorNodeDimensions(my_node[key], leaf_dims, dims, so_far = so_far+'|'+str(key))

            # Figure out the dimensions based on the layout schema
            # - make sure there's only one layout type for all the children
            child_layouts = set()
            for child in my_node.keys():
                child_layouts.add(self.__layoutType(child))
            if len(child_layouts) != 1:
                raise Exception(f'unknown layout mnemonic "{child_layouts}"')

            # - convert any of the children to the layout type
            child_layout = self.__layoutType(child)

            # - call the appropriate method to convert that layout to dimensions
            if   child_layout == 'borderLayout':
                leaf_dims[so_far] = self.__borderLayoutCalculation(my_node, leaf_dims, so_far)
            elif child_layout == 'gridLayout':
                leaf_dims[so_far] = self.__gridLayoutCalculation(my_node, leaf_dims, so_far)
            elif child_layout == 'flowLayoutHorizontal':
                leaf_dims[so_far] = self.__flowLayoutHorizontalCalculation(my_node, leaf_dims, so_far)
            elif child_layout == 'flowLayoutVertical':
                leaf_dims[so_far] = self.__flowLayoutVerticalCalculation(my_node, leaf_dims, so_far)
            else: # Unknown -- throw exception
                raise Exception(f'unknown layout mnemonic "{child}" / "{child_layout}"')

    #
    # Finalizize the widget placement
    #
    def __finalizeWidgetPlacement(self, my_node, leaf_dims, placement, x, y, w, h, h_gap, v_gap, widget_h_gap, widget_v_gap, so_far='root'):
        # Purely just the placement information
        placement[so_far] = (x,y,w,h)
        
        # Leaf node... which should be a widget
        if type(my_node) == tuple:
            placement[so_far+'|'+my_node[0]] = (x+widget_h_gap,y+widget_v_gap,w-2*widget_h_gap,h-2*widget_v_gap)

        # Leaf node... which is a tuple with only one element.. i.e., just the element
        elif type(my_node) == str:
            placement[so_far+'|'+my_node] = (x+widget_h_gap,y+widget_v_gap,w-2*widget_h_gap,h-2*widget_v_gap)
            
        # Interior node - requires processing the layout type and allocating spacing basd on the layout type 
        else:
            child_layout = self.__layoutType(next(iter(my_node.keys())))

            #
            # Border Layout
            #
            if   child_layout == 'borderLayout':
                dim_i = 0 # Use minimum dimensions by default
                if w < leaf_dims[so_far][1][0] and h < leaf_dims[so_far][1][1]:
                    dim_i = 1 # Okay to use preferred dimensions

                # Get the minimum (or preferred dimensions) for each cardinal tile
                c_dim = n_dim = s_dim = e_dim = w_dim = (0,0)
                if 'c' in my_node.keys():
                    c_dim = leaf_dims[so_far+'|c'][dim_i]
                if 'n' in my_node.keys():
                    n_dim = leaf_dims[so_far+'|n'][dim_i]
                if 's' in my_node.keys():
                    s_dim = leaf_dims[so_far+'|s'][dim_i]
                if 'e' in my_node.keys():
                    e_dim = leaf_dims[so_far+'|e'][dim_i]
                if 'w' in my_node.keys():
                    w_dim = leaf_dims[so_far+'|w'][dim_i]

                # The center gets any left overs... (if there's a center)
                if 'c' in my_node.keys():
                    # Give the center the left overs
                    c_dim = (w - (e_dim[0] + w_dim[0]), h - (s_dim[1] + n_dim[1]))

                    # Adjust everyone around the center...
                    if n_dim[0] > 0:
                        n_dim = (w, n_dim[1])
                    if s_dim[0] > 0:
                        s_dim = (w, s_dim[1])
                    if e_dim[1] > 0:
                        e_dim = (e_dim[0], c_dim[1])
                    if w_dim[1] > 0:
                        w_dim = (w_dim[0], c_dim[1])

                # Just north, just south, or just north and south? // no east, no west
                elif ('n' in my_node.keys() or 's' in my_node.keys()) and ('e' not in my_node.keys()) and ('w' not in my_node.keys()):
                    if    'n' in my_node.keys() and 's' in my_node.keys():
                        sum_h = leaf_dims[so_far + '|n'][dim_i][1] + leaf_dims[so_far + '|s'][dim_i][1]
                        n_dim = (w, h * n_dim[1]/sum_h)
                        s_dim = (w, h * s_dim[1]/sum_h)
                    elif  'n' in my_node.keys():
                        n_dim = (w, h)
                    else:
                        s_dim = (w, h)

                # Just east, just west, or just east and west? // no north, no south
                elif ('w' in my_node.keys() or 'e' in my_node.keys()) and ('n' not in my_node.keys()) and ('s' not in my_node.keys()):
                    if    'w' in my_node.keys() and 'e' in my_node.keys():
                        sum_w = leaf_dims[so_far + '|w'][dim_i][0] + leaf_dims[so_far + '|e'][dim_i][0]
                        w_dim = (w * w_dim[0]/sum_w, h)
                        e_dim = (w * e_dim[0]/sum_w, h)
                    elif  'w' in my_node.keys():
                        w_dim = (w, h)
                    else:
                        e_dim = (w, h)

                # have at least one in both directions...  everyone gets their minimum or preferred
                # -- left overs in height get allocated to the east and west
                # -- east and west (if they are both in existence, have to split the ratio
                else:
                    sum_w = 0
                    if 'w' in my_node.keys():
                        sum_w += leaf_dims[so_far + '|w'][dim_i][0]
                    if 'e' in my_node.keys():
                        sum_w += leaf_dims[so_far + '|e'][dim_i][0]

                    sum_h = 0
                    if 'n' in my_node.keys():
                        sum_h += leaf_dims[so_far + '|n'][dim_i][1]
                    if 's' in my_node.keys():
                        sum_h += leaf_dims[so_far + '|s'][dim_i][1]


                    if    'n' in my_node.keys() and 's' in my_node.keys():
                        n_dim = (w, n_dim[1])
                        s_dim = (w, s_dim[1])
                    elif  'n' in my_node.keys():
                        n_dim = (w, sum_h)
                    else:
                        s_dim = (w, sum_h)

                    if    'w' in my_node.keys() and 'e' in my_node.keys():
                        w_dim = (w * w_dim[0]/sum_w, h - sum_h)
                        e_dim = (w * e_dim[0]/sum_w, h - sum_h)
                    elif  'w' in my_node.keys():
                        w_dim = (w, h - sum_h)
                    else:
                        e_dim = (w, h - sum_h)


                if 'n' in my_node.keys():
                    my_x = x+h_gap
                    my_y = y+v_gap
                    self.__finalizeWidgetPlacement(my_node['n'], leaf_dims, placement, my_x, my_y, n_dim[0] - 2*h_gap, n_dim[1] - 2*v_gap, h_gap, v_gap, widget_h_gap, widget_v_gap, so_far + "|n")

                if 'w' in my_node.keys():
                    my_x = x + h_gap
                    my_y = y + n_dim[1] + v_gap
                    self.__finalizeWidgetPlacement(my_node['w'], leaf_dims, placement, my_x, my_y, w_dim[0] - 2*h_gap, w_dim[1] - 2*v_gap, h_gap, v_gap, widget_h_gap, widget_v_gap, so_far + "|w")

                if 'c' in my_node.keys():
                    my_x = x + w_dim[0] + h_gap
                    my_y = y + n_dim[1] + v_gap
                    self.__finalizeWidgetPlacement(my_node['c'], leaf_dims, placement, my_x, my_y, c_dim[0] - 2*h_gap, c_dim[1] - 2*v_gap, h_gap, v_gap, widget_h_gap, widget_v_gap, so_far + "|c")

                if 'e' in my_node.keys():
                    my_x = x + w_dim[0] + c_dim[0] + h_gap
                    my_y = y + n_dim[1] + v_gap
                    self.__finalizeWidgetPlacement(my_node['e'], leaf_dims, placement, my_x, my_y, e_dim[0] - 2*h_gap, e_dim[1] - 2*v_gap, h_gap, v_gap, widget_h_gap, widget_v_gap, so_far + "|e")

                if 's' in my_node.keys():
                    my_x = x
                    my_y = y + n_dim[1] + max(e_dim[1],c_dim[1],w_dim[1]) + v_gap
                    self.__finalizeWidgetPlacement(my_node['s'], leaf_dims, placement, my_x, my_y, s_dim[0] - 2*h_gap, s_dim[1] - 2*v_gap, h_gap, v_gap, widget_h_gap, widget_v_gap, so_far + "|s")

            #
            # Grid Layout
            #
            elif child_layout == 'gridLayout':
                # Determine the number of tiles in both x and y
                x_tiles = y_tiles = 1
                for key in my_node.keys():
                    if int(key.split(',')[0]) > x_tiles:
                        x_tiles = int(key.split(',')[0])
                    if int(key.split(',')[1]) > y_tiles:
                        y_tiles = int(key.split(',')[1])

                # Check the min dimensions... make it at least the mins...
                if w  < leaf_dims[so_far][0][0]:
                    w  = leaf_dims[so_far][0][0]
                if h < leaf_dims[so_far][0][1]:
                    h = leaf_dims[so_far][0][1]

                # Give each tile the same amount of space...
                tile_width  = w/x_tiles
                tile_height = h/y_tiles

                # Place the tiles and the subtree
                for x_tile in range(0,x_tiles):
                    for y_tile in range(0,y_tiles):
                        key = str(x_tile+1) + ',' + str(y_tile+1)
                        if key in my_node.keys():
                            self.__finalizeWidgetPlacement(my_node[key], leaf_dims, 
                                                           placement, 
                                                           x + x_tile*tile_width + h_gap, y + y_tile*tile_height + v_gap, 
                                                           tile_width-2*h_gap, tile_height-2*v_gap, 
                                                           h_gap, v_gap, widget_h_gap, widget_v_gap, so_far + "|" + key)

            #
            # Flow Layout (Horizontal)
            #
            elif child_layout == 'flowLayoutHorizontal':
                dim_i = 0 # use minimum dimensions
                if leaf_dims[so_far][1][0] < w:
                    dim_i = 1 # use preferred dimensions

                # determine the sum of dimension needed (for ratio)
                sum_w   = 0
                to_sort = []
                for key in my_node.keys():
                    to_sort.append(int(key))
                    sum_w += leaf_dims[so_far + "|" + key][dim_i][0]

                to_sort = sorted(to_sort)
                x_inc = x
                for i in range(len(to_sort)-1,-1,-1):
                    key = so_far + "|" + str(to_sort[i])
                    w_ratio = leaf_dims[key][dim_i][0] / sum_w
                    self.__finalizeWidgetPlacement(my_node[str(to_sort[i])], leaf_dims, placement, x_inc + h_gap, y + v_gap, w_ratio*w - 2*h_gap, h - 2*v_gap, 
                                                   h_gap, v_gap, widget_h_gap, widget_v_gap, key)
                    x_inc += w_ratio*w

            #
            # Flow Layout (Vertical)
            # - MOSTLY A COPY OF FLOW LAYOUT (HORIZONTAL) WITH TRANSPOSE
            #
            elif child_layout == 'flowLayoutVertical':
                dim_i = 0 # use minimum dimensions
                if leaf_dims[so_far][1][1] < h:
                    dim_i = 1 # use preferred dimensions

                # determine the sum of dimension needed (for ratio)
                sum_h   = 0
                to_sort = []
                for key in my_node.keys():
                    to_sort.append(int(key))
                    sum_h += leaf_dims[so_far + "|" + key][dim_i][1]

                to_sort = sorted(to_sort)
                y_inc = y
                for i in range(0,len(to_sort)):
                    key = so_far + "|" + str(to_sort[i])
                    h_ratio = leaf_dims[key][dim_i][1] / sum_h
                    self.__finalizeWidgetPlacement(my_node[str(to_sort[i])], leaf_dims, placement, x + h_gap, y_inc + v_gap, w - 2*h_gap, h_ratio*h - 2*v_gap, 
                                                   h_gap, v_gap, widget_h_gap, widget_v_gap, key)
                    y_inc += h_ratio*h

            #
            # Error...
            #
            else:
                raise Exception(f'unknown layout mnemonic "{my_node}" / "{child_layout}"')

    #
    # Create a spatial dimension lookup based on a dashboard specification (as a dictionary)
    #
    def dictionaryLayoutToSpatialDimensions(self, spec, w, h, h_gap=0, v_gap=0, widget_h_gap=1, widget_v_gap=1):
        # Obtain the minimum and preferred dimensions dynamically
        dims = {}
        for widget in self.widgets:
            dims[widget] = {}
            dims[widget]['pref'] = eval(f'self.{widget}PreferredDimensions()')
            dims[widget]['min']  = eval(f'self.{widget}MinimumDimensions()')
        
        # Obtain the minimum and preferred dimensions dynamically // new version passes params to the preferred dimension methods
        for k,v in spec.items():
            widget = v
            if len(v) == 2:
                widget        = v[0]
            widget_params = {}
            if len(v) == 2:
                widget_params = v[1]
            dims[str(v)]         = {}
            dims[str(v)]['pref'] = eval(f'self.{widget}PreferredDimensions(**widget_params)')
            dims[str(v)]['min']  = eval(f'self.{widget}MinimumDimensions(**widget_params)')
        
        # Create the layout tree -- fill in the leaves with the widget dimensions
        # - layout_tree[layout_mnemonic] -> [child_layout_mnemonic] -> ... -> widget description
        layout_tree = {}
        for key in spec.keys():
            value = spec[key]
            self.__recursivePlace(layout_tree, key, value)

        # Calculate the minimum and preferred size for the interior nodes of the tree
        # - leaf_dims[<original_keys_from_the_spec>] = ((minimum dimensions - x,y),(preferred dimensions - x,y))
        leaf_dims = {}
        self.__fillInteriorNodeDimensions(layout_tree,leaf_dims,dims)

        # Finalize the placement -- reconcile either the preferred (requested) with the minimum widget size
        placement = {}
        self.__finalizeWidgetPlacement(layout_tree, leaf_dims, placement, 0, 0, w, h, h_gap, v_gap, widget_h_gap, widget_v_gap)

        # return placement,layout_tree,leaf_dims # Used for debugging the internal datastructures
        return placement

    #
    # Create SVG for the placement data structure
    # - To debug the placement really...
    #
    def placementSVGDebug(self, placement):
        # Start the svg object
        svg =  f'<svg width="{placement["root"][2]+1}" height="{placement["root"][3]+1}" xmlns="http://www.w3.org/2000/svg">'

        for key in placement.keys():
            # Ignore the widgets themselves... i.e., just show the placement positions
            render = True
            for widget in self.widgets:
                if widget in key:
                    render = False
            if render:
                coords = placement[key]
                svg += f'<rect width="{coords[2]}" height="{coords[3]}" x="{coords[0]}" y="{coords[1]}" stroke="#000000" fill-opacity="0.0" />'
                svg += f'<text x="{coords[0] + coords[2]/2}" y="{coords[1] + coords[3]/2}" text-anchor="middle">{key}</text>'

        # Close out the svg object
        svg += '</svg>'
        return svg
    
    #
    # Generic layout method -- will choose proper layout method...
    #
    def layout(self,
               spec,                                # Multiwidget specification
               df,                                  # Dataframe to render
               #------------------------------------#
               widget_id        = None,             # Widget ID
               #------------------------------------#
               w                = 1024,             # Width of the multi-widget panel
               h                = 1024,             # Height of the multi-widget panel
               h_gap            = 0,                # Horizontal left/right gap
               v_gap            = 0,                # Verticate top/bottom gap
               widget_h_gap     = 1,                # Horizontal gap between widgets
               widget_v_gap     = 1,                # Vertical gap between widgets
               track_state      = False,            # Track state of geometry to data frame // for Panel
               rt_reactive_html = None,             # Reference to Reactive HTML Panel // for Panel
               **kwargs):
        # Determine type of specification...
        str_count,tup_count,unk_count = 0,0,0
        for k in spec.keys():
            if type(k) == str:
                str_count += 1
            elif type(k) == tuple:
                tup_count += 1
            else:
                unk_count += 1
        if    str_count >  0 and tup_count == 0 and unk_count == 0:
            return self.multiWidgetPanel(df,spec,widget_id,w,h,h_gap,v_gap,widget_h_gap,widget_v_gap,track_state,rt_reactive_html,**kwargs)
        elif  str_count == 0 and tup_count >  0 and unk_count == 0:
            return self.gridBagLayout   (df,spec,widget_id,w,h,h_gap,v_gap,widget_h_gap,widget_v_gap,track_state,rt_reactive_html,**kwargs)
        else:
            raise Exception(f'rt.layout() failed to recognize specification type {str_count}/{tup_count}/{unk_count}')

    #
    # Create the SVG multipanel widget based on the spec, the dataframe, and the dynamic variables
    #
    def multiWidgetPanel(self,
                         df,                                  # Dataframe to render
                         spec,                                # Multiwidget specification
                         #------------------------------------#
                         widget_id        = None,             # Widget ID
                         #------------------------------------#
                         w                = 1024,             # Width of the multi-widget panel
                         h                = 1024,             # Height of the multi-widget panel
                         h_gap            = 0,                # Horizontal left/right gap
                         v_gap            = 0,                # Verticate top/bottom gap
                         widget_h_gap     = 1,                # Horizontal gap between widgets
                         widget_v_gap     = 1,                # Vertical gap between widgets
                         track_state      = False,            # Track state of geometry to data frame // for Panel 
                         rt_reactive_html = None,             # Reference to Reactive HTML Panel // for Panel
                         **kwargs):
        # Widget ID
        if widget_id is None:
            widget_id = 'multiwidget_panel_' + str(random.randint(0,65535))

        # Calculate the placement
        placement = self.dictionaryLayoutToSpatialDimensions(spec, w, h, h_gap, v_gap, widget_h_gap, widget_v_gap)

        # General application calculations
        # ... time granularity
        widgets_set = set()
        for k,v in spec.items():
            if type(v) == tuple:
                widgets_set.add(v[0])
            else:
                widgets_set.add(v)

        if ('temporal_granularity' not in kwargs.keys()) and 'temporalBarChart' in widgets_set:
            if ('ts_field' not in kwargs.keys()) or kwargs['ts_field'] is None:
                ts_field = self.guessTimestampField(df)
            else:
                ts_field = kwargs['ts_field']
            temporal_granularity = self.temporalGranularity(df, ts_field)

        # Go through the placement, identify the placement keys that are widgets... fill them in... then render them...
        _instance_lu = {}
        for place in placement.keys():
            # Only keep the widgets
            render = False
            for widget in self.widgets:
                if widget in place:
                    render = True
                    
            # For the widgets... create the custom svg call and add to the svg element
            if render:
                spec_key   = place[place.index('|')+1:place.rindex('|')]
                spec_tuple = spec[spec_key]
                widget_loc = placement[place]
                
                # Create the custom params for the widget creation
                my_params = kwargs.copy()
                my_params['df']          = df
                my_params['x_view']      = widget_loc[0]
                my_params['y_view']      = widget_loc[1]
                my_params['w']           = widget_loc[2]
                my_params['h']           = widget_loc[3]
                my_params['track_state'] = track_state
                
                # If the spec_tuple is actually a tuple, they copy those parts into the params as well
                # - these should override any passed from this method (the kwargs.copy())
                if type(spec_tuple) == tuple:
                    widget_method = spec_tuple[0]
                    for k in spec_tuple[1].keys():
                        v = spec_tuple[1][k]
                        my_params[k] = v
                else:
                    widget_method = spec_tuple

                # Need to remove any args that aren't applicable to this widget
                accepted_args = set(inspect.getfullargspec(getattr(self, widget_method)).args)
                to_remove = set()
                for k in my_params.keys():
                    if k not in accepted_args:
                        to_remove.add(k)
                for x in to_remove:
                    my_params.pop(x)

                # General application parameters
                if 'temporal_granularity' in accepted_args and 'temporal_granularity' not in my_params:
                    my_params['temporal_granularity'] = temporal_granularity
                if 'rt_reactive_html' in accepted_args:
                    my_params['rt_reactive_html'] = rt_reactive_html

                # Resolve the method name and invoke it adding to the svg string
                _func      = getattr(self, widget_method)
                _instance  = _func(**my_params)

                _px0,_py0,_px1,_py1 = my_params['x_view'], my_params['y_view'], my_params['x_view']+my_params['w'], my_params['y_view']+my_params['h']
                _instance_bounds = Polygon([[_px0,_py0],[_px0,_py1],[_px1,_py1],[_px1,_py0]])
                _instance_lu[_instance_bounds] = _instance

        return RTComponentsLayout(self, widget_id, _instance_lu, w, h)

    #
    # Create the SVG multipanel widget using a gridbag-like layout method.
    # ... spec --> spec[x,y,w,h - tuple] -> ('component',{'param1':'value1', ... })
    #
    def gridBagLayout(self,
                      df,                                    # Dataframe to render
                      spec,                                  # Multiwidget specification
                      #--------------------------------------#
                      widget_id        = None,               # Widget ID
                      #--------------------------------------#
                      w                = 1024,               # Width of the multi-widget panel
                      h                = 1024,               # Height of the multi-widget panel
                      h_gap            = 0,                  # Horizontal left/right gap
                      v_gap            = 0,                  # Verticate top/bottom gap
                      widget_h_gap     = 1,                  # Horizontal gap between widgets
                      widget_v_gap     = 1,                  # Vertical gap between widgets
                      track_state      = False,              # Track the state of the dataframe to geometry // for Panel
                      rt_reactive_html = None,               # Reference to Reactive HTML Panel // for Panel
                      **kwargs):
        # Widget ID
        if widget_id is None:
            widget_id = 'panel_' + str(random.randint(0,65535))
        
        # Validate the layout first
        # - Determine the max coordinate
        widgets_set = set()
        _tile_x_max,_tile_y_max = 1,1
        for xywh_tuple in spec.keys():
            _x0,_y0,_tile_w,_tile_h = xywh_tuple
            if (_x0 + _tile_w) > _tile_x_max:
                _tile_x_max = _x0 + _tile_w
            if (_y0 + _tile_h) > _tile_y_max:
                _tile_y_max = _y0 + _tile_h
            if type(spec[xywh_tuple]) == tuple:
                widgets_set.add(spec[xywh_tuple][0])
            else:
                widgets_set.add(spec[xywh_tuple])
        # - Allocate a representation
        cells = [[0 for i in range(_tile_x_max)] for j in range(_tile_y_max)] # cells[y][x]
        for _y in range(_tile_y_max):
            for _x in range(_tile_x_max):
                cells[_y][_x] = -1
        # - Place tuples into representation - checking to make sure there isn't overlap
        for xywh_tuple in spec.keys():
            _x0,_y0,_tile_w,_tile_h = xywh_tuple
            for _x in range(_x0,_x0+_tile_w):
                for _y in range(_y0,_y0+_tile_h):
                    if cells[_y][_x] != -1:
                        raise Exception(f'gridBagLayout() - overlapping coordinate @ {_x},{_y}')
                    cells[_y][_x] = xywh_tuple
        
        # Determine the tile size -- make it into actual pixels... x-pixels-per-tile (xppt)
        _xppt,_yppt = ceil((w-2*h_gap)/_tile_x_max),ceil((h-2*v_gap)/_tile_y_max)
        w,h         = _xppt * _tile_x_max, _yppt * _tile_y_max

        # Determine temporal granularity if there's a temporalBarChart...
        if ('temporal_granularity' not in kwargs.keys()) and 'temporalBarChart' in widgets_set:
            if ('ts_field' not in kwargs.keys()) or kwargs['ts_field'] is None:
                ts_field = self.guessTimestampField(df)                
            else:
                ts_field = kwargs['ts_field']
            temporal_granularity = self.temporalGranularity(df, ts_field)
        else:
            temporal_granularity = None

        # Start the SVG
        _instance_lu = {}
        for xywh_tuple in spec.keys():
            _x0,_y0,_tile_w,_tile_h = xywh_tuple
            spec_tuple = spec[xywh_tuple]
                
            # Create the custom params for the widget creation
            my_params = kwargs.copy()
            my_params['df']          = df
            my_params['x_view']      = h_gap + _x0 * _xppt + widget_h_gap
            my_params['y_view']      = v_gap + _y0 * _yppt + widget_v_gap
            my_params['w']           = _tile_w * _xppt - 2 * widget_h_gap
            my_params['h']           = _tile_h * _yppt - 2 * widget_v_gap
            my_params['track_state'] = track_state
                
            # If the spec_tuple is actually a tuple, then copy those parts into the params as well
            # - these should override any passed from this method (the kwargs.copy())
            if type(spec_tuple) == tuple:
                widget_method = spec_tuple[0]
                for k in spec_tuple[1].keys():
                    v = spec_tuple[1][k]
                    my_params[k] = v
            else:
                widget_method = spec_tuple

            # Need to remove any args that aren't applicable to this widget
            accepted_args = set(inspect.getfullargspec(getattr(self, widget_method)).args)
            to_remove = set()
            for k in my_params.keys():
                if k not in accepted_args:
                    to_remove.add(k)
            for x in to_remove:
                my_params.pop(x)

            # General application parameters
            if 'temporal_granularity' in accepted_args and 'temporal_granularity' not in my_params:
                my_params['temporal_granularity'] = temporal_granularity
            if 'rt_reactive_html' in accepted_args:
                my_params['rt_reactive_html'] = rt_reactive_html

            # Resolve the method name and invoke it adding to the svg string
            _func      = getattr(self, widget_method)
            _instance  = _func(**my_params)
            
            _px0,_py0,_px1,_py1 = my_params['x_view'], my_params['y_view'], my_params['x_view']+my_params['w'], my_params['y_view']+my_params['h']
            _instance_bounds = Polygon([[_px0,_py0],[_px0,_py1],[_px1,_py1],[_px1,_py0]])
            _instance_lu[_instance_bounds] = _instance

        return RTComponentsLayout(self, widget_id, _instance_lu, w, h)

    #
    # panelControlPreferredDimensions() - preferred dimensions for the panel control panel
    #
    def panelControlPreferredDimensions(self, **kwargs):
        return (256, 24)

    #
    # panelControlMinimumDimensions() - minimum dimensions for the panel control panel
    #
    def panelControlMinimumDimensions(self, **kwargs):
        return (64, 20)

    #
    # panelControl() - return instance of the panel control panel
    #
    def panelControl(self,
                     df=None,
                     rt_reactive_html=None,
                     txt_h=None,
                     x_view=0,
                     y_view=0,
                     x_ins=2,
                     y_ins=2,
                     w=128,
                     h=20):
        _params_ = locals().copy()
        _params_.pop('self')
        return self.RTPanelControl(self, **_params_)

    #
    # RTPanelControl() - return instance of the panel control panel
    #
    class RTPanelControl(RTComponent):
        #
        # Constructor
        #
        def __init__(self, rt_self, **kwargs):
            self.rt_self          = rt_self
            self.df               = kwargs['df']
            self.x_view           = kwargs['x_view']
            self.y_view           = kwargs['y_view']
            self.w                = kwargs['w']
            self.h                = kwargs['h']
            self.x_ins            = kwargs['x_ins']
            self.y_ins            = kwargs['y_ins']
            self.rt_reactive_html = kwargs['rt_reactive_html']
            self.last_render      = None                         # For Compatibility w/ RTComponent Usage
            self.geom_to_df       = {}

            if 'txt_h' in kwargs.keys() and kwargs['txt_h'] is not None:
                self.txt_h = kwargs['txt_h']
            else:
                self.txt_h = self.h - 2 * self.y_ins
                if self.txt_h > 24:
                    self.txt_h = 24

        #
        # _repr_svg_() - SVG Representation
        # - always re-render (do not keep the last render around)
        # - this method is unique across the RTComponent class since last_render is never used
        #
        def _repr_svg_(self):
            return self.renderSVG()
        
        #
        # renderSVG() - render the SVG representation of the panel control
        #
        def renderSVG(self):
            self.geom_to_df = {}
            _svg_ = f'<svg id="panel_control" x="{self.x_view}" y="{self.y_view}" width="{self.w}" height="{self.h}" xmlns="http://www.w3.org/2000/svg">'
            bg_color = self.rt_self.co_mgr.getTVColor('background','default')
            _svg_ += f'<rect x="0" y="0" width="{self.w}" height="{self.h}" fill="{bg_color}" stroke="{bg_color}" stroke-width="1"/>'
            _svg_ += self.rt_self.svgText('RT', self.x_ins, self.y_ins + self.txt_h, self.txt_h)
            rt_l   = self.rt_self.textLength('RT', self.txt_h) + 2 * self.x_ins
            # If an RT Reactive HTML component was specified, draw information about the state of the panel
            if self.rt_reactive_html is not None:
                # Blocks to indicate the depth
                _depth_     = self.rt_reactive_html.df_level + 1
                _blocks_    = len(self.rt_reactive_html.dfs)
                block_color = self.rt_self.co_mgr.getTVColor('data','default')
                block_gap   = 2
                block_w     = (self.h - 2 * self.y_ins) if self.h < 24 else 20
                for xi in range(_blocks_):
                    _fill_   = block_color if xi < _depth_ else 'none'
                    x = self.x_ins + rt_l + xi * (block_w + block_gap)
                    _svg_ += f'<rect x="{x}" y="{self.y_ins}" width="{block_w}" height="{block_w}" fill="{_fill_}" stroke="{block_color}" stroke-width="1"/>'
                    _poly_ = Polygon([[x, self.y_ins],[x, self.y_ins+block_w],[x+block_w, self.y_ins+block_w],[x+block_w, self.y_ins]])
                    self.geom_to_df[_poly_] = self.rt_reactive_html.dfs[xi]
                # Text string to indicate the size of the dataframe
                if self.df is not None:
                    _readable_ = self.rt_self.readable(len(self.df))
                    _svg_ += self.rt_self.svgText(_readable_, self.x_ins + rt_l + _blocks_*(block_w+block_gap) + block_gap, self.y_ins + self.txt_h, self.txt_h)
            _svg_ += '</svg>'
            return _svg_

        #
        # overlappingDataFrames() - Determine which dataframe geometris overlap with a specific one
        #
        def overlappingDataFrames(self, to_intersect):
            _dfs = []
            for _poly in self.geom_to_df.keys():
                if _poly.intersects(to_intersect):
                    _dfs.append(self.geom_to_df[_poly])
            if len(_dfs) > 0:
                return _dfs[0]
            else:
                return None

#
# RT Layout Instance (new version)
#
class RTComponentsLayout(RTComponent):
    #
    # Constructor
    #
    def __init__(self, rt_self, widget_id, instance_lu, w, h):
        self.rt_self     = rt_self
        self.widget_id   = widget_id
        self.instance_lu = instance_lu
        self.w           = w
        self.h           = h
        self.last_render = None
    
    #
    # SVG Representation
    #
    def _repr_svg_(self):
        if self.last_render is None:
            self.renderSVG()
        return self.last_render
    
    #
    # renderSVG() - render the SVG representation
    #
    def renderSVG(self):
        _svg_ = [f'<svg id="{self.widget_id}" width="{self.w+1}" height="{self.h+1}" xmlns="http://www.w3.org/2000/svg">']
        for _poly in self.instance_lu.keys():
            _svg_.append(self.instance_lu[_poly]._repr_svg_())
        _svg_.append('</svg>')
        self.last_render = ''.join(_svg_)
        return self.last_render

    #
    # applyViewConfigurations() - apply view configurations based on a reference layout.
    #
    def applyViewConfigurations(self, ref_layout):
        for _poly_ in self.instance_lu.keys():
            self.instance_lu[_poly_].applyViewConfiguration(ref_layout.instance_lu[_poly_])

    #
    # identifyComponent() - identify the component that the coordinate falls within
    # - return the component, the key for this component, and the modified/relative coordinate adjusted to the component
    #
    def identifyComponent(self, coordinate):
        for _poly_ in self.instance_lu.keys():
            bnds = _poly_.bounds
            if coordinate[0] >= bnds[0] and coordinate[0] <= bnds[2] and coordinate[1] >= bnds[1] and coordinate[1] <= bnds[3]:
                return self.instance_lu[_poly_], _poly_, (coordinate[0] - bnds[0], coordinate[1] - bnds[1])
        return None, None, None

    #
    # componentInstance() - return the component instance for the specified key
    # - used in conjuction with the key returned in identifyComponent()
    # - and should work across layers in the interactive stack
    #
    def componentInstance(self, k):
        return self.instance_lu[k] if k in self.instance_lu.keys() else None

    #
    # entityPositions() - return information about the entity geometry for rendering
    def entityPositions(self, entity):
        all_positions = []
        for _poly_ in self.instance_lu.keys():
            _instance_  = self.instance_lu[_poly_]
            _positions_ = _instance_.entityPositions(entity)
            if _positions_ is not None and len(_positions_) > 0:
                for _pos_ in _positions_:
                    _pos_.xyOffset((_poly_.bounds[0], _poly_.bounds[1]))
                    all_positions.append(_pos_)
        return all_positions

    #
    # overlappingDataFrames() - Return Overlapping Dataframes
    #
    def overlappingDataFrames(self, to_intersect):
        if type(to_intersect) == tuple: # assume it's a bounding box (x0,y0,x1,y1)
            _x0,_y0,_x1,_y1 = to_intersect
            if _x0 > _x1:
                _x0,_x1 = _x1,_x0
            if _y0 > _y1:
                _y0,_y1 = _y1,_y0
            to_intersect = Polygon([[_x0,_y0],[_x0,_y1],[_x1,_y1],[_x1,_y0]])
        _dfs = []
        for _poly in self.instance_lu.keys():
            if _poly.intersects(to_intersect):
                adj_to_intersect = translate(to_intersect, -_poly.bounds[0], -_poly.bounds[1])
                _df = self.instance_lu[_poly].overlappingDataFrames(adj_to_intersect)
                if _df is not None and len(_df) > 0:
                    _dfs.append(_df)
        if len(_dfs) > 0:
            return self.rt_self.concatDataFrames(_dfs)
        else:
            return None
